open Misc
open Parsetree
open Exp
open Symbtree

exception Undef_non_pure_expression 
exception Conversion_error of string * Location.t

let mkst desc loc =
    { can_st_desc = desc;
      can_st_lv = IdSet.empty;
      can_st_loc = loc; }

let (>>) f g x = g (f x)

(* Option monad *)
let (>>=) x f =
  match x with
  | None -> None
  | Some v -> f v

(** {2 Conversion to canonical form} *)

let rec p_exp_to_can_exp e = match e.pexp_desc with
  | Pexp_ident s -> Ce_ident (create_ident s)
  | Pexp_num n -> Ce_num n
  | Pexp_bool b -> Ce_num (if b then 1 else 0)
  | Pexp_prefix (op,e1) -> 
      let ce1 = p_exp_to_can_exp e1 in 
      begin match op with
	| "+" -> ce1
	| "-" -> mk_grp("+",0,[(-1,ce1)])
	| _ -> mk_fun(op,[ce1])
      end
  | Pexp_infix (op,e1,e2) -> 
      let ce1 = p_exp_to_can_exp e1 in 
      let ce2 = p_exp_to_can_exp e2 in 
      if Pervasives.(=) op "+" || Pervasives.(=) op "^" then mk_grp(op,0,[(1,ce1);(1,ce2)])
      else if Pervasives.(=) op "-" then mk_grp("+",0,[(1,ce1);(-1,ce2)]) 
      else mk_fun (op,[ce1;ce2])
  | Pexp_cast(e1,_) -> p_exp_to_can_exp e1
  | Pexp_fun(i,el) -> mk_fun(i, List.map p_exp_to_can_exp el)
  | Pexp_fcall _
  | Pexp_new _
  | Pexp_fld(_,_) -> raise Undef_non_pure_expression

(** [p_exp_to_can_exp_se exp statements] returns the statements
 required to evaluate [exp] appended to [statements] and a canonical
 expression ({!Symbtree.can_exp}) *)
let rec p_exp_to_can_exp_se e stmts = match e.pexp_desc with
  | Pexp_ident s -> (stmts, Ce_ident (create_ident s))
  | Pexp_num n -> (stmts, Ce_num n)
  | Pexp_bool b -> (stmts, Ce_num (if b then 1 else 0))
  | Pexp_prefix (op,e1) -> 
      let (stmts, ce1) = p_exp_to_can_exp_se e1 stmts in 
      let ce = match op with
	| "+" -> ce1
	| "-" -> mk_grp("+",0,[(-1,ce1)])
	| _ -> mk_fun(op,[ce1])
      in (stmts, ce)
  | Pexp_infix (op,e1,e2) -> 
      let (stmts, ce1) = p_exp_to_can_exp_se e1 stmts in 
      let (stmts, ce2) = p_exp_to_can_exp_se e2 stmts in 
      (stmts, if Pervasives.(=) op "+" || Pervasives.(=) op "^" then mk_grp(op,0,[(1,ce1);(1,ce2)])
              else if Pervasives.(=) op "-" then mk_grp("+",0,[(1,ce1);(-1,ce2)]) 
	      else mk_fun (op,[ce1;ce2]))
  | Pexp_cast(e1,_) -> p_exp_to_can_exp_se e1 stmts
  | Pexp_fun(i,el) -> 
      let (stmts, cel) = p_exp_list_to_can_exp_list_se el stmts in 
      (stmts, mk_fun(i, cel))
  | Pexp_fcall(s,(idl,el)) -> 
      let id = tempid () in
      let idl = List.map (fst >> create_ident) idl in
      let (stmts2, cel) = p_exp_list_to_can_exp_list_se el [] in
      (stmts2 @ (mkst (Cst_pfcall[(id,s,idl,cel)]) e.pexp_loc)::stmts, Ce_ident id)
  | Pexp_new _ ->
      let id = tempid() in  
      ((mkst (Cst_new(id)) e.pexp_loc)::stmts, Ce_ident id)
  | Pexp_fld (e,c) ->
      let id = tempid () in 
      let (stmts2, ce) = p_exp_to_can_exp_se e [] in 
      (stmts2 @ (mkst (Cst_fldlookup(id, ce, c)) e.pexp_loc)::stmts, Ce_ident id)

and p_exp_list_to_can_exp_list_se el stmts = 
  List.fold_left 
    (fun (stmts,cel) -> fun ae -> 
      let (stmts, ce) = p_exp_to_can_exp_se ae stmts in
      (stmts,ce::cel))
    (stmts,[])
    (List.rev el)

let can_prop_atom ca = can_prop_pure (pure_one ca)

let p_exp_to_can_prop e =
  let rec go e = match e.pexp_desc with
    | Pexp_prefix ("!",e1) ->
	let (cp1,cp2) = go e1 in
        (cp2,cp1)
    | Pexp_infix ("&&",e1,e2) -> 
	let (cp11,cp12) = go e1 in
	let (cp21,cp22) = go e2 in
	(can_prop_star cp11 cp21, can_prop_or cp12 (can_prop_star cp11 cp22))
    | Pexp_infix ("||",e1,e2) ->
	let (cp11,cp12) = go e1 in
	let (cp21,cp22) = go e2 in
	(can_prop_or cp11 (can_prop_star cp12 cp21), can_prop_star cp12 cp22)
    | Pexp_infix ((("<"|"<="|">"|">="|"=="|"!=") as op),e1,e2) -> 
      let (e1,e2) = (p_exp_to_can_exp e1, p_exp_to_can_exp e2) in
      let ca = match op with 
	| "<"  -> mk_POS(-1,[(-1,e1);(1,e2)])
	| "<=" -> mk_POS(0,[(-1,e1);(1,e2)])
	| ">"  -> mk_POS(-1,[(1,e1);(-1,e2)])
	| ">=" -> mk_POS(0,[(1,e1);(-1,e2)])
	| "==" -> mk_EQ(e1,e2)
	| "!=" -> mk_NEQ(e1,e2)
	| _ -> assert false in
      (can_prop_atom ca, can_prop_atom (neg_can_atom ca))
    | _ -> 
      let e = p_exp_to_can_exp e in
      (can_prop_atom (mk_NEQ(Ce_num 0,e)), can_prop_atom (mk_EQ(Ce_num 0,e)))
  in go e 

let p_exp_to_can_prop_se e = 
  let rec go e = match e.pexp_desc with
    | Pexp_prefix ("!",e1) ->
	let (stmts1,stmts2) = go e1 in
        (stmts2,stmts1)
    | Pexp_infix ("&&",e1,e2) -> 
      let (stmts11,stmts12) = go e1 in 
      let (stmts21,stmts22) = go e2 in 
      (stmts11 @ stmts21, [mkst (Cst_nondet(stmts12,stmts11 @ stmts22)) e.pexp_loc] )
    | Pexp_infix ("||",e1,e2) ->
      let (stmts11,stmts12) = go e1 in 
      let (stmts21,stmts22) = go e2 in 
      ([mkst (Cst_nondet(stmts11,stmts12 @ stmts21)) e.pexp_loc], stmts12 @ stmts22)
    | Pexp_infix ((("<"|"<="|">"|">="|"=="|"!=") as op),e1,e2) -> 
      let (stmts,e1) = p_exp_to_can_exp_se e1 [] in 
      let (stmts,e2) = p_exp_to_can_exp_se e2 stmts in 
      let ca = match op with 
	| "<"  -> mk_POS(-1,[(-1,e1);(1,e2)])
	| "<=" -> mk_POS(0,[(-1,e1);(1,e2)])
	| ">"  -> mk_POS(-1,[(1,e1);(-1,e2)])
	| ">=" -> mk_POS(0,[(1,e1);(-1,e2)])
	| "==" -> mk_EQ(e1,e2)
	| "!=" -> mk_NEQ(e1,e2)
	| _ -> assert false in
      let cond1 = [mkst (Cst_assume(can_prop_atom ca)) e.pexp_loc] in
      let cond2 = [mkst (Cst_assume(can_prop_atom (neg_can_atom ca))) e.pexp_loc] in
      (stmts @ cond1, stmts @ cond2)
    | _ ->
      let (stmts,ce) = p_exp_to_can_exp_se e [] in
      let cond1 = [mkst (Cst_assume(can_prop_atom (mk_NEQ(Ce_num 0,ce)))) e.pexp_loc] in
      let cond2 = [mkst (Cst_assume(can_prop_atom (mk_EQ(Ce_num 0,ce)))) e.pexp_loc] in
      (stmts @ cond1, stmts @ cond2)
  in go e

(* Argument list is simply a node name. *)
let read_node argl loc = match argl with
  | []      -> "." ^ Config.list_link_tag
  | [(snode,loc)] ->
      let _ = Inductive_preds.add_ip_use (snode,-1,loc) in
      snode 
  | _::_::_ -> raise (Conversion_error ("Syntax error: Bad arguments to predicate.", loc))

let a_space_pred_to_can_spred = function
  | Aspred_node (s,e,cel) ->
      let f (c,e) = (c, p_exp_to_can_exp e) in
      Csp_node (s,p_exp_to_can_exp e, fld_from_list (List.map f cel))
  | Aspred_indpred ("list",argl,el,loc) ->
      let snode = read_node argl loc in
      let (e,sfun,f) = match el with
	| [e] -> (p_exp_to_can_exp e, "@zero", Ce_num 0)
	| [e;f] -> (p_exp_to_can_exp e, !Config.value_abstraction, p_exp_to_can_exp f)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `list' predicate.", loc)) in
      Csp_listsegi(SINGLE(snode,sfun,fld_from_list []), e, Ce_num 0, f, Ce_num 0, dset_emp)
  | Aspred_indpred (("lseg"|"lsegi" as x),argl,el,loc) ->
      let snode = read_node argl loc in
      let (e,f,sfun,g) = match el with
	| [e;f] -> (p_exp_to_can_exp e, p_exp_to_can_exp f, "@zero", Ce_num 0)
	| [e;f;g] -> (p_exp_to_can_exp e, p_exp_to_can_exp f, !Config.value_abstraction, p_exp_to_can_exp g)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `lseg' predicate.", loc)) in
      Csp_listsegi(SINGLE(snode,sfun,fld_from_list []), e, f, g, Ce_num 0, 
		   if x="lseg" then dset_from_list [f] else dset_emp)
  | Aspred_indpred ("dlseg",argl,el,loc) ->
      let (t1,t2) = match argl with
	| []      -> (Config.dl_Rlink_tag, Config.dl_Llink_tag)
	| [(t1,_);(t2,_)] -> (t1,t2)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `dlseg' predicate.", loc)) in
      let (e1,f1,e2,f2) = match el with
	| [e1;f1;e2;f2] -> (p_exp_to_can_exp e1, p_exp_to_can_exp f1, p_exp_to_can_exp e2, p_exp_to_can_exp f2)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `dlseg' predicate.", loc)) in
      Csp_listsegi(DOUBLE(t1,t2), e1, f1, e2, f2, dset_from_list [f1; e2])
  | Aspred_indpred ("xlseg",argl,el,loc) ->
      let t1 = match argl with
	| []       -> (Config.dl_Llink_tag)
	| [(t1,_)] -> t1
	| _::_::_ -> raise (Conversion_error ("Syntax error: Bad arguments to `xlseg' predicate.", loc)) in
      let (e1,f1,e2,f2) = match el with
	| [e1;f1;e2;f2] -> (p_exp_to_can_exp e1, p_exp_to_can_exp f1, p_exp_to_can_exp e2, p_exp_to_can_exp f2)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `xlseg' predicate.", loc)) in
      Csp_listsegi(XOR t1, e1, f1, e2, f2, dset_from_list [f1; e2])
  | Aspred_indpred ("tree",argl,el,loc) ->
      let (t1,t2) = match argl with
	| []      -> Config.tree_link_tags
	| [(t1,_);(t2,_)] -> (t1,t2)
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `tree' predicate.", loc)) in
      let e = match el with
	| [e] -> p_exp_to_can_exp e
	| _ -> raise (Conversion_error ("Syntax error: Bad arguments to `tree' predicate.", loc)) in
      Csp_tree (t1, t2, e)
  | Aspred_indpred (i,[],el,loc) ->
      let _ = Inductive_preds.add_ip_use (i,List.length el,loc) in
      Csp_indpred (i, List.map p_exp_to_can_exp el)
  | Aspred_indpred (_,_::_,_,loc) ->
      raise (Conversion_error ("Syntax error: Bad arguments to predicate.", loc))

(*
  | Aspred_list (snode,e,loc) ->
      let _ = Inductive_preds.add_ip_use (snode,-1,loc) in
      Csp_listsegi(SINGLE(snode,"@zero",fld_from_list []), p_exp_to_can_exp e, Ce_num 0, Ce_num 0,
		   Ce_num 0, (* this argument is garbage *)
		   dset_emp)
  | Aspred_listseg (snode,sfun,e1,e2,e3,el,loc) ->
      let _ = Inductive_preds.add_ip_use (snode,-1,loc) in
      Csp_listsegi(SINGLE(snode,sfun,fld_from_list []), p_exp_to_can_exp e1, p_exp_to_can_exp e2,
		   p_exp_to_can_exp e3,
		   Ce_num 0, (* this argument is garbage *)
		   dset_from_list (p_exp_to_can_exp e2 :: List.map p_exp_to_can_exp el))
  | Aspred_listsegi (snode,sfun,e1,e2,e3,el,loc) ->
      let _ = Inductive_preds.add_ip_use (snode,-1,loc) in
      Csp_listsegi(SINGLE(snode,sfun,fld_from_list []), p_exp_to_can_exp e1, p_exp_to_can_exp e2,
		   p_exp_to_can_exp e3,
		   Ce_num 0, (* this argument is garbage *)
		   dset_from_list (List.map p_exp_to_can_exp el))
  | Aspred_dlseg (k,t1,e1,f1,t2,e2,f2) ->
      Csp_listsegi((if Pervasives.(=) k DL then DOUBLE(t1,t2) else XOR t1),
                   p_exp_to_can_exp e1, p_exp_to_can_exp f1,
                   p_exp_to_can_exp e2, p_exp_to_can_exp f2,
		   dset_from_list [p_exp_to_can_exp f1;p_exp_to_can_exp e2])
  | Aspred_node (s,e,cel) ->
      let f (c,e) = (c, p_exp_to_can_exp e) in
      Csp_node (s,p_exp_to_can_exp e, fld_from_list (List.map f cel))
  | Aspred_tree (t1,t2,e)->
      Csp_tree (t1,t2, p_exp_to_can_exp e)
  | Aspred_indpred (i,el,loc) ->
      let _ = Inductive_preds.add_ip_use (i,List.length el,loc) in
      Csp_indpred (i, List.map p_exp_to_can_exp el)
*)

let prop_to_can_prop p loc = 
  let rec go p = match p with
  | Aprop_exp e -> fst (p_exp_to_can_prop e)
  | Aprop_spred sp -> can_prop_spred (spat_one (a_space_pred_to_can_spred sp))
  | Aprop_ifthenelse (e,p1,p2) ->
      let (cond1,cond2) = p_exp_to_can_prop e in
      let px = can_prop_star cond1 (go p1) in
      let py = can_prop_star cond2 (go p2) in
      can_prop_or px py
  | Aprop_star (p1,p2) -> can_prop_star (go p1) (go p2)
  | Aprop_barbar (p2,p3) -> can_prop_or (go p2) (go p3)
  | Aprop_wande (p1,p2) ->
      (try can_prop_wande (go p1) (go p2)
      with Undef_wand_elim s -> raise (Conversion_error ("Could not eliminate -*E: " ^ s, loc)))
  | Aprop_box (rid,p) -> {cfl = [((pure_true,spat_empty),[(rid,normalize_can_prop (go p))])]}
  in
    normalize_can_prop (go p)

let prop_to_ip_body p loc =
  let join (pl1,cf1) (pl2,cf2) = 
    normalize_pure (pure_conj pl1 pl2) >>= fun pl -> 
    can_form_star cf1 cf2 >>= fun cf ->
    Some (pl,cf) in
  let joinl xl1 xl2 = List.concat (List.map (fun x -> map_partial (join x) xl2) xl1) in
  let rec go = function
  | Aprop_exp e -> 
      begin match (fst (p_exp_to_can_prop e)).cfl with 
	| [] -> []
	| [((pl,_),[])] -> [(pure_true,((pl,spat_empty),[]))]
	| _ -> assert false
      end
  | Aprop_spred sp -> [(pure_true,((pure_true,spat_one (a_space_pred_to_can_spred sp)),[]))]
  | Aprop_ifthenelse (e,p1,p2) ->
      let (cond1,cond2) = p_exp_to_can_prop e in
      let p1 = match cond1.cfl with
	| [] -> []
	| [((pl,_),[])] -> List.map (fun(guard,cf)->(pure_conj pl guard,cf)) (go p1)
	| _ -> assert false in
      let p2 = match cond2.cfl with
	| [] -> []
	| [((pl,_),[])] -> List.map (fun(guard,cf)->(pure_conj pl guard,cf)) (go p2)
	| _ -> assert false in
      p1 @ p2
  | Aprop_star (p1,p2) -> joinl (go p1) (go p2)
  | Aprop_barbar (p2,p3) -> raise (Conversion_error ("Syntax error: Disjunction (||) is not allowed in user-defined predicates.", loc))
  | Aprop_wande (p1,p2) -> raise (Conversion_error ("Syntax error: Septraction (-*E) is not allowed in user-defined predicates.", loc))
  | Aprop_box (rid,p) -> [(pure_true,((pure_true,spat_empty),[(rid,prop_to_can_prop p loc)]))]
  in
  go p
