(** Abstract syntax tree for expressions and pure formulae *)

open Misc

(** {2 Expressions} *)

(** Canonical expressions: 
    Do not construct [Ce_fun]s and [Ce_grp]s directly, but rather use the
    functions [mk_fun] and [mk_grp]. 
 *)
type can_exp =
  | Ce_num of int
  | Ce_ident of Misc.ident
  | Ce_fun of string * can_exp list
  | Ce_grp of string * int * (int * can_exp) list (* abelian groups *)

val mk_fun : string * can_exp list -> can_exp
val mk_XOR : can_exp * can_exp -> can_exp
val mk_grp : string * int * (int * can_exp) list -> can_exp

val equal_exp : can_exp -> can_exp -> bool
val compare_exp : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_exp -> can_exp -> int
val compare_exp_list : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_exp list -> can_exp list -> int

val mem_exp : can_exp -> can_exp list -> bool

val remove_exp_duplicates : can_exp list -> can_exp list

(** return true if can_exp of the form _x *)
val existential_id : can_exp -> bool
val fv_can_exp : can_exp -> Misc.IdSet.t
val ex_fv_can_exp : can_exp -> Misc.IdSet.t


(** {2 Atomic pure formulae} *)

(** Canonical atom: equality or inequality *)
type can_atom 

val mk_EQ  : can_exp * can_exp -> can_atom
val mk_NEQ : can_exp * can_exp -> can_atom
val mk_POS : int * (int * can_exp) list -> can_atom

val is_EQ: can_atom -> bool
val is_NEQ: can_atom -> bool
val un_EQ: can_atom -> can_exp * can_exp
val un_NEQ: can_atom -> can_exp * can_exp

val compare_atom : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_atom -> can_atom -> int
val fold_atom : (can_exp -> 'a -> 'a) -> can_atom -> 'a -> 'a

(** {2 Pure formulae} *)

(** Conjunction of canonical atoms *)
type can_pure

val pure_true : can_pure
val pure_one : can_atom -> can_pure
val pure_conj_one : can_atom -> can_pure -> can_pure 
val pure_conj : can_pure -> can_pure -> can_pure
val pure_fold : (can_atom -> 'a -> 'a) -> can_pure -> 'a -> 'a
val pure_to_list : can_pure -> can_atom list
val pure_is_true : can_pure -> bool
val pure_filter : (can_atom -> bool) -> can_pure -> can_pure
val pure_length : can_pure -> int

val mk_EQ_exn : can_exp * can_exp -> can_pure option
val mk_NEQ_exn : can_exp * can_exp -> can_pure option
val is_uneq : can_pure -> can_exp * can_exp -> bool

val mem_atom : can_atom -> can_pure -> bool

val compare_pure : (*?cmp_id:(Misc.ident -> Misc.ident -> int) ->*) can_pure -> can_pure -> int

val remove_atom_duplicates : can_pure -> can_pure
val remove_pure_duplicates : can_pure list -> can_pure list

val normalize_can_atom : can_atom -> can_pure option
val normalize_pure : can_pure -> can_pure option

val neg_can_atom : can_atom -> can_atom
val eval_can_atom : can_pure -> can_atom -> bool


(** Check whether [pl |- ca]. Returns [Some true] if [pl => ca];
   [Some false] if [pl => not ca]; [None] if uncertain. *)
val entails_atom : can_pure -> can_atom -> bool option 

(** [entails_pure_list pl pl'] returns 
       [None]  if [pl /\ pl' <=> false]; 
    or [Some pl''] if [pl /\ pl'' ==> pl']. *)
val entails_pure_list : can_pure -> can_pure -> can_pure option 

val fv_norm_can_atom : can_atom -> Misc.IdSet.t
val fv_keep_can_atom : can_atom -> Misc.IdSet.t
val fv_can_atom : can_atom -> Misc.IdSet.t


(** {2 Disjoint sets} *)

(** Disjoint sets *)
type disj_set

val dset_emp: disj_set
val dset_add: can_exp -> disj_set -> disj_set
val dset_remove: can_exp -> disj_set -> disj_set
val dset_mem: can_exp -> disj_set -> bool
val dset_from_list: can_exp list -> disj_set
val dset_to_list: disj_set -> can_exp list 
val dset_inter: disj_set -> disj_set -> disj_set
val dset_subset: disj_set -> disj_set -> bool
val dset_partition: (can_exp -> bool) -> disj_set -> can_exp list * disj_set
val dset_notmem: can_pure -> can_exp -> disj_set -> bool

(** Returns the set of expresssions that are not equal to e in pl *)
val get_uneql : can_exp -> can_pure -> disj_set
val fv_dset : disj_set -> Misc.IdSet.t

val compare_dset : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> disj_set -> disj_set -> int

(** {2 Fields} *)  

(** Structure field names *)
type component = string

(** Partial map from components to expressions *)
type fields

val fld_from_list : (component * can_exp) list -> fields
val fld_to_list : fields -> (component * can_exp) list
val fld_inter : fields -> fields -> fields
val fld_union : fields -> fields -> fields
val fld_get : component -> fields -> can_exp
val fld_set : component * can_exp -> fields -> fields
val fld_mem : component * can_exp -> fields -> bool
val fld_containing : can_exp -> fields -> component
val fld_exists : (component * can_exp -> bool) -> fields -> bool
val fld_filter : (component * can_exp -> bool) -> fields -> fields
val fld_remove : component -> fields -> fields
val fld_subset : (Misc.ident * can_exp) list -> fields -> fields -> ((Misc.ident * can_exp) list * can_pure) option
val weak_fld_subset : (Misc.ident * can_exp) list -> fields -> fields -> ((Misc.ident * can_exp) list * can_pure) option

val components_equal : fields -> fields -> can_pure option
val getcomponent : component -> fields -> can_exp option

val compare_fields : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> fields -> fields -> int
val mem_component_exp : component * can_exp -> (component * can_exp) list -> bool

val fv_fld : fields -> Misc.IdSet.t

(** {2 Substitutions} *)

type subst = can_exp -> can_exp
val mk_subst: (Misc.ident * can_exp) list -> subst
val mk_single_subst : (Misc.ident * can_exp) -> subst
val mk_gensym_garb_subst : Misc.ident -> subst
val mk_gensym_subst_list : Misc.ident list -> subst
val mk_gensym_garb_subst_list : Misc.ident list -> subst

val map_atom : (can_exp -> can_exp) -> can_atom -> can_atom
val map_pure : (can_exp -> can_exp) -> can_pure -> can_pure
val map_dset : (can_exp -> can_exp) -> disj_set -> disj_set
val map_fld  : (can_exp -> can_exp) -> fields -> fields

(** {2 Equalities} *)

type equalities
val empty_eq : equalities

(** extend equalities eq with e1=e2 *)
val eq_add : equalities -> (can_exp * can_exp) -> equalities option

(** extend equalities eq with e1=f1...en=fn *)
val eq_add_list : equalities -> can_pure -> equalities option
val eq_partition_list : equalities -> can_pure -> (equalities * can_pure) option

val sub_equalities : subst -> equalities -> equalities
val eq_to_sub : equalities -> subst
val eq_to_can_pure : equalities -> can_pure

val combine_eq : equalities -> equalities -> (equalities * equalities * equalities)

(* remove binding for e from eq *)
val kill_vars_eq : equalities -> Misc.IdSet.t -> (equalities * subst)

(** {2 Constraints} *)

type constr = (can_exp * can_exp) list * (can_exp * can_exp)
type constraints
val empty_constrs : constraints
val add_constr : constraints -> constr -> constraints
val sub_constrs : subst -> constraints -> constraints
val simplify_constraints : can_pure * constraints -> can_pure * constraints
val solve_constraints : equalities -> can_pure -> constraints -> (equalities * can_pure * constraints) option

(** {2 Misc operations} *)


(** Replace existentials with gensym'd normal vars. *)
val existentials_rename_tonormal_sub : Misc.IdSet.t -> subst

(** Replace existentials with gensym'd existential vars. *)
val existentials_rename_sub : Misc.IdSet.t -> subst

val weak_unify_exp : (Misc.ident * can_exp) list -> can_exp -> can_exp
	-> ((Misc.ident * can_exp) list * can_pure) option

(** use equalities in [pl] and try to derive a contradiction *) 
val aggr_norm_pure : can_pure -> can_pure option

(** normalize list and remove equalities *)
val aggr_prove_pure : can_pure -> can_pure option
 

(** {2 Pretty printing} *)

val pp_listsep : (Format.formatter -> 'a -> unit) ->
      string -> string -> Format.formatter -> 'a list -> unit
val pp_can_exp : Format.formatter -> can_exp -> unit
val pp_can_atom : Format.formatter -> can_atom -> unit
val pp_eq : Format.formatter -> equalities -> unit
val pp_constraints : Format.formatter -> constraints -> unit

