type pipe_str =
    { p2c_in : in_channel;
      p2c_out : out_channel;
      c2p_in : in_channel;
      c2p_out : out_channel;
      mutable child_id : int }

let send_to_child p_str v =
  Marshal.to_channel p_str.p2c_out v [];
  flush p_str.p2c_out

let receive_from_child p_str =
  Marshal.from_channel p_str.c2p_in

let receive_from_parent p_str =
  match Marshal.from_channel p_str.p2c_in with
    | Some n ->
        n
    | None ->
	(* Format.printf "@._%d_ Received stop signal, now stopping%!@." p_str.child_id; *)
	exit(0)

let send_to_parent p_str n =
  Marshal.to_channel p_str.c2p_out n [];
  flush p_str.c2p_out

let service_f n = n+1

let num_of_processes = ref 0

let incr_process_count () =
  incr num_of_processes
  (* Format.printf "@.Number of processes: %d@." !num_of_processes *)

let decr_process_count () =
  decr num_of_processes
  (* Format.printf "@.Number of processes: %d@." !num_of_processes *)

let spawn_fun service_f =
  let p_str = 
    let (p2c_read,p2c_write) = Unix.pipe () in
    let (c2p_read,c2p_write) = Unix.pipe () in
      (* Unix.set_nonblock c2p_read; *)
      { p2c_in = Unix.in_channel_of_descr p2c_read;
	p2c_out = Unix.out_channel_of_descr p2c_write;
	c2p_in = Unix.in_channel_of_descr c2p_read;
	c2p_out = Unix.out_channel_of_descr c2p_write;
        child_id = 0 }
  in match Unix.fork () with
    | 0 ->
        p_str.child_id <- Unix.getpid ();
	let rec loop () =
	  let n = receive_from_parent p_str in
	  let res = service_f n in
	    send_to_parent p_str res;
	    loop ()
	in loop ()
    | cid ->
	p_str.child_id <- cid;
	incr_process_count ();
	p_str

let stop_child p_str =
  send_to_child p_str None;
  ignore (Unix.waitpid [] p_str.child_id);
  Unix.close (Unix.descr_of_in_channel p_str.p2c_in);
  Unix.close (Unix.descr_of_out_channel p_str.p2c_out);
  Unix.close (Unix.descr_of_in_channel p_str.c2p_in);
  Unix.close (Unix.descr_of_out_channel p_str.c2p_out);
  decr_process_count ()

(*
let _ =
  let p_str1 = spawn_fun service_f in
  let p_str2 = spawn_fun service_f in
    send_to_child p_str1 (Some 7);
    send_to_child p_str2 (Some 11);
    let res1 = receive_from_child p_str1 in
    let res2 = receive_from_child p_str2
    in send_to_child p_str1 None;
      send_to_child p_str2 None
*)

let par_map num_processes (f: 'a -> 'b) (l:'a list) =
  let a = Array.init num_processes (fun _ -> spawn_fun f) in
  let avail_list = ref (Array.to_list a) in
  let busy_list = ref [] in
  let input_l : 'a list ref = ref l in
  let output_l : 'b list ref = ref [] in
    while (!input_l <> [] || !busy_list <> []) do
      if !input_l <> [] && !avail_list <> [] then
	let p_str = List.hd !avail_list in
	  send_to_child p_str (Some (List.hd !input_l));
	  avail_list := List.tl !avail_list;
	  busy_list := p_str :: !busy_list;
	  input_l := List.tl !input_l
      else
	let rec f = function
	  | p_str::l ->
	      (try
		  let v = receive_from_child p_str in
		    output_l := v :: !output_l;
		    avail_list := p_str :: !avail_list;
		    f l
		with Sys_blocked_io -> p_str :: f l)
	  | [] -> []
	in busy_list := f !busy_list
    done;
    List.iter stop_child !avail_list;
    !output_l

let nest_level = ref 0
let max_nest_level = 1

let parallel_map ?(num_processes=2) f l =
 if List.length l >= 1 && !nest_level < max_nest_level && num_processes>1 then
   begin
     incr nest_level;
     Unix.handle_unix_error (par_map num_processes f) l
   end
 else List.map f l

(*
let _ =
  let f x =
    begin
      for i = 1 to 1000000 do () done;
      let _ = raise Not_found
      in ()
    end in
  let _ = parallel_map f (Array.to_list  (Array.init 3 (fun x -> x)))
  in Format.printf "@.Finished@."
*)
