%{ (* header *)

open Misc
open Parsetree

let mkexp d = { pexp_desc = d; pexp_loc = Location.symbol_loc() }
let mkexp_ghost d = { pexp_desc = d; pexp_loc = Location.none }
let mkstm d = { pstm_desc = d; pstm_loc = Location.symbol_loc() }
let mkstm_ghost d = { pstm_desc = d; pstm_loc = Location.none }

let mk_ref_params cel loc =
  let check_par = function
    | {pexp_desc = Pexp_ident i; pexp_loc=l} -> (i,l)
    | _ -> raise(Location.Parse_error("Syntax error: Reference parameters must be variables.", loc)) in
  List.map check_par cel
  
let mk_indpred_params cel =
  let check_par = function
    | {pexp_desc = Pexp_ident i; pexp_loc=l} -> (i,l)
    | {pexp_loc = l} -> raise(Location.Parse_error("Syntax error: Node parameters must be variables.", l)) in
  List.map check_par cel
 
(* implicitly called when no grammar rules apply *)
let parse_error _ =
  raise(
    Location.Parse_error("Syntax error.",
      match !Location.lexbuf with
	| None -> Location.symbol_loc()
	| Some lexbuf ->
	    (* the Parsing library only updates symbol_end_pos when successfully
	     * reducing a grammar rule, so here we ask the lexer for the current
	     * position directly *)
	    Location.mkloc (Parsing.symbol_start_pos()) lexbuf.Lexing.lex_curr_p))


%} /* declarations */

/* tokens */

%token ACTION
%token AMPERAMPER
%token AS
%token BARBAR
%token <bool> BOOL
%token BREAK
%token CLASS
%token COLON
%token COMMA
%token CONSTRUCTOR
%token CONTINUE
%token DISPOSE
%token DO
%token DOT
%token ELSE
%token EMPTY
%token ENSURES
%token EOF
%token EQUAL
%token <string> IDENT
%token <string> QIDENT
%token <string> FIDENT
%token IF
%token <string> INFIXOP1
%token <string> INFIXOP2
%token <string> INFIXOP3
%token INVARIANT
%token LBRACE
%token LBRACKET
%token LET
%token LPAREN
%token MINUSGREATER
%token <int> NAT
%token NEW
%token PAR
%token POINTSTO
%token RBRACE
%token RBRACKET
%token REQUIRES
%token RESOURCE
%token RETURN
%token RPAREN
%token SEMI
%token STAR
%token THEN
%token TREE
%token <string> UNARYOP
%token UNDERSCORE
%token VOID
%token WHEN
%token WANDEXISTS
%token WHILE
%token WITH

/* precedences (increasing) and associativities for expressions */

%nonassoc below_ELSE
%nonassoc ELSE AS 
%right BARBAR
%right WANDEXISTS
%left STAR AMPERAMPER
%left POINTSTO
%left INFIXOP1 EQUAL
%left INFIXOP2
%left INFIXOP3
%nonassoc UNARYOP
%left MINUSGREATER

/* entry points */

%start program
%type <Parsetree.p_program> program

%start a_proposition
%type <Parsetree.a_proposition> a_proposition


%% /* rules */

/* entry points */
program:
  | program_item_star      { List.rev $1 }
;
program_item_star:
  | /* empty */                     { [] }
  | program_item_star var_decls     { $2 @ $1 }
  | program_item_star class_decl    { $2 :: $1 }
  | program_item_star fun_decl      { $2 :: $1 }
  | program_item_star resource_decl { $2 :: $1 }
  | program_item_star indpred_decl  { $2 :: $1 }
;
var_decls:
  | IDENT ident_notempty_seq SEMI { List.rev_map (fun (x,y)-> Pdec_var(x,$1,y)) $2 }
;
class_decl:
  | CLASS IDENT LBRACE field_decls RBRACE  { Pdec_class($2, List.rev $4, Location.rhs_loc 2) }
;
field_decls:
  | /* empty */                { [] }
  | field_decls IDENT IDENT SEMI  { ($3,$2,Location.rhs_loc 3)::$1 }
;
fun_decl:
  | VOID IDENT LPAREN formal_params RPAREN LBRACE top_statements RBRACE
      { Pdec_fun($2,"void",$4,(None,None),$7,Location.symbol_loc()) }
  | VOID IDENT LPAREN formal_params RPAREN fun_specs LBRACE top_statements RBRACE
      { Pdec_fun($2,"void",$4,$6,$8,Location.symbol_loc()) }
  | fun_specs VOID IDENT LPAREN formal_params RPAREN LBRACE top_statements RBRACE
      { Pdec_fun($3,"void",$5,$1,$8,Location.symbol_loc()) }
  | IDENT IDENT LPAREN formal_params RPAREN LBRACE top_statements RBRACE
      { Pdec_fun($2,$1,$4,(None,None),$7,Location.symbol_loc()) }
  | IDENT IDENT LPAREN formal_params RPAREN fun_specs LBRACE top_statements RBRACE
      { Pdec_fun($2,$1,$4,$6,$8,Location.symbol_loc()) }
  | fun_specs IDENT IDENT LPAREN formal_params RPAREN LBRACE top_statements RBRACE
      { Pdec_fun($3,$2,$5,$1,$8,Location.symbol_loc()) }
;
fun_specs:
  | REQUIRES a_proposition  { (Some $2, None) }
  | ENSURES a_proposition   { (None, Some $2) }
  | REQUIRES a_proposition ENSURES a_proposition   { (Some $2, Some $4) }
;
resource_decl:
  | RESOURCE IDENT LBRACE invariant_decl constructor_decl action_seq RBRACE
      { Pdec_resource($2,[],$4,$5,$6,Location.symbol_loc()) }
  | RESOURCE IDENT LPAREN ident_seq RPAREN LBRACE invariant_decl constructor_decl action_seq RBRACE
      { Pdec_resource($2,$4,$7,$8,$9,Location.symbol_loc()) }
;
invariant_decl:
  | /* empty */              { None }
  | INVARIANT a_proposition  { Some $2 }
;
constructor_decl:
  | /* empty */              { ([], []) }
  | CONSTRUCTOR LBRACE top_statements RBRACE  { $3 }
;
action_decl:
  | ACTION IDENT LPAREN ident_seq RPAREN LBRACKET a_proposition RBRACKET
	LBRACKET a_proposition RBRACKET
      { ($2,$4,a_prop_empty,$7,$10,Location.symbol_loc()) }
  | ACTION IDENT LPAREN ident_seq RPAREN LBRACKET a_proposition RBRACKET
	LBRACKET a_proposition RBRACKET LBRACKET a_proposition RBRACKET
      { ($2,$4,$7,$10,$13,Location.symbol_loc()) }
;
action_seq:
  | /* empty */			{ [] }
  | action_decl action_seq	{ $1::$2 }
;
indpred_decl:
  | LET IDENT LPAREN ident_seq RPAREN EQUAL LBRACKET a_proposition RBRACKET
      { Pdec_indpred($2,$4,$8,Location.symbol_loc()) }
;
ident_seq:
  | /* empty */      { [] }
  | ident_notempty_seq { $1 }
;
ident_notempty_seq:
  | IDENT            { [($1,Location.rhs_loc 1)] }
  | IDENT COMMA ident_notempty_seq { ($1,Location.rhs_loc 1)::$3 }
;
tyid_seq:
  | /* empty */       { [] }
  | tyid_notempty_seq { $1 }
;
tyid_notempty_seq:
  | IDENT IDENT      { [($2,$1,Location.rhs_loc 2)] }
  | IDENT IDENT COMMA tyid_notempty_seq  { ($2,$1,Location.rhs_loc 2)::$4 }
;
top_statements:
  | /* empty */              { ([],[]) }
  | IDENT ident_notempty_seq SEMI top_statements { (List.map (fun (x,y) -> (x,$1,y)) $2 @ fst $4, snd $4) }
  | statement statement_star { ([], $1::$2) }
;
statement_star:
  | /* empty */              { [] }
  | statement statement_star { $1::$2 }
;
statement:
  | expression MINUSGREATER IDENT EQUAL expression SEMI
	  { mkstm(Pstm_fldassign($1, $3, $5)) }
  | IDENT EQUAL expression SEMI
	  { mkstm(Pstm_assign($1, $3)) }
  | DISPOSE expression SEMI
          { mkstm(Pstm_dispose($2)) }
  | LBRACE statement_star RBRACE
	  { mkstm(Pstm_block($2)) }
  | IF LPAREN expression RPAREN statement %prec below_ELSE
	  { mkstm(Pstm_if(Some $3, $5, mkstm_ghost(Pstm_block []))) }
  | IF LPAREN expression RPAREN statement ELSE statement
	  { mkstm(Pstm_if(Some $3, $5, $7)) }
  | IF LPAREN STAR RPAREN statement %prec below_ELSE
	  { mkstm(Pstm_if(None, $5, mkstm_ghost(Pstm_block []))) }
  | IF LPAREN STAR RPAREN statement ELSE statement
	  { mkstm(Pstm_if(None, $5, $7)) }
  | WHILE LPAREN expression RPAREN invariant statement
	  { mkstm(Pstm_while($5, Some $3, $6)) }
  | WHILE LPAREN STAR RPAREN invariant statement
	  { mkstm(Pstm_while($5, None, $6)) }
  | WITH IDENT optional_when statement %prec below_ELSE
          { mkstm(Pstm_withres($2,$3,$4,"",[])) }
  | WITH IDENT optional_when statement AS IDENT LPAREN expression_seq RPAREN SEMI
          { mkstm(Pstm_withres($2,$3,$4,$6,$8)) }
  | DO statement AS IDENT DOT IDENT LPAREN expression_seq RPAREN SEMI
          { mkstm(Pstm_action($2,$4,$6,$8)) }
  | expression SEMI               { mkstm(Pstm_exp($1)) }
  | PAR LBRACE proc_calls RBRACE  { mkstm(Pstm_parblock(List.rev $3)) }
  | RETURN SEMI                   { mkstm(Pstm_return(None)) }
  | RETURN expression SEMI        { mkstm(Pstm_return(Some $2)) }
  | BREAK SEMI                    { mkstm(Pstm_break) }
  | CONTINUE SEMI                 { mkstm(Pstm_continue) }
;
proc_call:
  | IDENT LPAREN actual_params RPAREN SEMI              { (None,$1,$3) }
  | IDENT EQUAL IDENT LPAREN actual_params RPAREN SEMI  { (Some $1,$3,$5) }
;
proc_calls:
  | proc_call              { [$1] }
  | proc_calls proc_call   { $2::$1 }
;
optional_when:
  | /* empty */                   { mkexp(Pexp_infix("==", mkexp_ghost(Pexp_num 0), mkexp_ghost(Pexp_num 0))) }
  | WHEN LPAREN expression RPAREN { $3 }
;
invariant:
  | /* empty */                     { None }
  | LBRACKET a_proposition RBRACKET { Some $2 }
;
expression:
  | IDENT   { mkexp(Pexp_ident($1)) }
  | NAT     { mkexp(Pexp_num($1)) }
  | BOOL    { mkexp(Pexp_bool $1) }
  | LPAREN expression RPAREN          { $2 }
  | LPAREN expression AS IDENT RPAREN { mkexp(Pexp_cast($2,$4)) }
  | UNARYOP expression                { mkexp(Pexp_prefix($1, $2)) }
  | INFIXOP2 expression %prec UNARYOP { mkexp(Pexp_prefix($1, $2)) }
  | expression MINUSGREATER IDENT     { mkexp(Pexp_fld($1, $3)) }
  | NEW LPAREN RPAREN                 { mkexp(Pexp_new "any") }
  | NEW IDENT LPAREN RPAREN           { mkexp(Pexp_new $2) }
  | expression BARBAR expression      { mkexp(Pexp_infix("||", $1, $3)) }
  | expression AMPERAMPER expression  { mkexp(Pexp_infix("&&", $1, $3)) }
  | expression STAR expression        { mkexp(Pexp_infix("*", $1, $3)) }
  | expression INFIXOP1 expression    { mkexp(Pexp_infix($2, $1, $3)) }
  | expression INFIXOP2 expression    { mkexp(Pexp_infix($2, $1, $3)) }
  | expression INFIXOP3 expression    { mkexp(Pexp_infix($2, $1, $3)) }
  | IDENT LPAREN actual_params RPAREN { mkexp(Pexp_fcall($1, $3)) }
  | FIDENT LPAREN expression_seq RPAREN { mkexp(Pexp_fun($1, $3)) }
;
expression_seq:
  | /* empty */      { [] }
  | expression_notempty_seq { $1 }
;
expression_notempty_seq:
  | expression            { [$1] }
  | expression COMMA expression_notempty_seq { $1::$3 }
;
a_expression_seq:
  | /* empty */      { [] }
  | a_expression_notempty_seq { $1 }
;
a_expression_notempty_seq:
  | a_expression            { [$1] }
  | a_expression COMMA a_expression_notempty_seq { $1::$3 }
;
formal_params:
  | tyid_seq               { ([],$1) }
  | tyid_seq SEMI tyid_seq { ($1,$3) }
;
actual_params:
  | expression_seq { ([],$1) }
  | expression_seq SEMI expression_seq { (mk_ref_params $1 (Location.rhs_loc 1), $3) }
;
a_component_expression_seq:
  | UNDERSCORE                          { [] }
  | a_component_expression_notempty_seq { $1 }
;
a_component_expression_notempty_seq:
  | IDENT COLON a_expression            { [($1,$3)] }
  | IDENT COLON a_expression COMMA a_component_expression_notempty_seq { ($1,$3)::$5 }
;
a_space_pred:
  | IDENT LPAREN a_expression_seq RPAREN                          { Aspred_indpred($1,[],$3,Location.symbol_loc()) }
  | IDENT LPAREN a_expression_seq SEMI a_expression_seq RPAREN    { Aspred_indpred($1,mk_indpred_params $3,$5,Location.symbol_loc()) }
  | a_expression POINTSTO a_component_expression_seq          { Aspred_node("Node",$1,$3) }
  | a_expression POINTSTO a_expression                        { Aspred_node("Node",$1,[(Config.list_link_tag, $3)]) }
/*  | a_expression POINTSTO a_expression COMMA a_expression     { Aspred_node("Node",$1,[(fst Config.tree_link_tags, $3); (snd Config.tree_link_tags, $5)]) }
*/
;
a_proposition:
  | LPAREN a_proposition RPAREN { $2 }
  | EMPTY                       { a_prop_empty }
  | a_expression                { Aprop_exp $1}
  | a_space_pred                { Aprop_spred $1 }
  | IF a_expression THEN a_proposition ELSE a_proposition { Aprop_ifthenelse($2,$4,$6) }
  | a_proposition STAR a_proposition            { Aprop_star($1,$3) }
  | a_proposition BARBAR a_proposition          { Aprop_barbar($1,$3) }
  | a_proposition WANDEXISTS a_proposition      { Aprop_wande($1,$3) }
  | IDENT COLON LBRACKET a_proposition RBRACKET { Aprop_box ($1,$4) }
;
a_expression:
  | LPAREN a_expression RPAREN            { $2 }
  | IDENT                                 { mkexp(Pexp_ident($1))       }
  | QIDENT                                { mkexp(Pexp_ident($1))       }
  | NAT                                   { mkexp(Pexp_num($1))         }
  | BOOL                                  { mkexp(Pexp_bool($1))        }
  | FIDENT LPAREN a_expression_seq RPAREN { mkexp(Pexp_fun($1,$3))      }
  | UNARYOP expression                    { mkexp(Pexp_prefix($1, $2)) }
  | INFIXOP2 expression %prec UNARYOP     { mkexp(Pexp_prefix($1, $2)) }
  | a_expression AMPERAMPER a_expression  { mkexp(Pexp_infix("&&",$1,$3)) }
  | a_expression INFIXOP1 a_expression    { mkexp(Pexp_infix($2,$1,$3)) }
  | a_expression INFIXOP2 a_expression    { mkexp(Pexp_infix($2,$1,$3)) }
  | a_expression INFIXOP3 a_expression    { mkexp(Pexp_infix($2,$1,$3)) }
;

%% (* trailer *)
