open Misc
open Exp
open Symbtree
open Inductive_preds
open Format

(* Option monad *)
let (>>=) x f = 
  match x with 
  | None -> None
  | Some v -> f v

let un_Some xo =
  match xo with
  | None -> assert false
  | Some x -> x

let with_default y x = 
  match x with 
  | None -> y
  | Some v -> v

exception MonadInternal

let lifted_fold_left f e xl =
  let rec go e xl = match (e,xl) with
    | (None,_) -> None
    | (e,[]) -> e
    | (Some e,x::xl) -> go (f e x) xl in
  go e xl

let lifted_map f xl =
  let rec go = function
    | [] -> []
    | x::xl -> match f x  with
               | None -> raise MonadInternal
               | Some y -> y :: go xl
  in try
    Some (go xl)
  with MonadInternal -> None 


let split_spatial filter sl =
  match spat_fold (fun x res -> match res with None -> filter x >>= fun y -> Some(x,y) | Some _ -> res) sl None with
  | None -> raise Not_found
  | Some (sp,y) -> (y, spat_remove sl sp)

let node_filter nid e = function
  | Csp_node(nid',e',el') when Pervasives.(=) nid' nid && equal_exp e e' -> Some el'
  | _ -> None

let using_induction () =
  if !Config.show_induction then
  fprintf !Config.formatter "USING INDUCTION@."

(* base id name for unrolling of predicates *)
let split_id_base = create_ident "split"

(** unroll spatial predicate csp to expose epto|-> *)
let unroll_can_sp pl epto csp = match csp with
  | Csp_node (_,e,_) -> if equal_exp e epto then Some (pure_true,spat_one csp) else None
  | Csp_listsegi (SINGLE (snode,sfun,cel),e1,f1,e2,f2,el) ->
      if is_uneq pl (e1,f1) && equal_exp e1 epto
      then
	let (e3,e4,nid,cel',pl') = unroll_node_lhs snode e1 in
        let e5 = Ce_ident (gensym (create_ident "myval")) in
	mk_EQ_exn (e2, mk_fun(sfun,[e4;e5])) >>= fun pl'' ->
	Some (pure_conj pl'' pl',
	      spat_star_one (Csp_node(nid,e1, fld_union cel' cel))
	        (spat_one (Csp_listsegi (SINGLE(snode,sfun,cel),e3,f1,e5,f2,el))))
      else None  
  | Csp_listsegi (k,e1,f1,e2,f2,el) ->
      if is_uneq pl (e1,f1) && equal_exp e1 epto  (* unroll from lhs *)
      then
        let split_id = gensym split_id_base in
        let cel = match k with 
          | SINGLE _ -> assert false
          | DOUBLE(t1,t2) -> fld_from_list [(t1,Ce_ident split_id);(t2,e2)]
          | XOR(t1) -> fld_from_list [(t1,mk_XOR(Ce_ident(split_id),e2))] in
        Some (pure_true,
	      spat_star_one (Csp_node ("Node",e1, cel))
	        (spat_one (Csp_listsegi (k,Ce_ident split_id, f1,e1,f2,el))))
      else if is_uneq pl (e2,f2) && equal_exp f2 epto  (* unroll from rhs *)
      then
        let _ = using_induction () in
        let split_id = gensym split_id_base in
        let cel = match k with 
          | SINGLE _ -> assert false
          | DOUBLE(t1,t2) -> fld_from_list [(t1,f1);(t2,Ce_ident split_id)] 
          | XOR(t1) -> fld_from_list [(t1,mk_XOR(Ce_ident(split_id),f1))] in
        Some (pure_true,
	      spat_star_one (Csp_node ("Node",f2, cel))
		(spat_one (Csp_listsegi (k,e1,f2,e2,Ce_ident split_id,el))))
      else None
  | Csp_tree (t1,t2,ce) ->
      if is_uneq pl (Ce_num 0,ce) && equal_exp ce epto
      then
	let split_id1 = gensym split_id_base in
	let split_id2 = gensym split_id_base in
	let cel = fld_from_list [(t1,Ce_ident split_id1);(t2,Ce_ident split_id2)] in
	Some (pure_true,
	      spat_star_one (Csp_node ("Node",ce, cel))
		(spat_star_one (Csp_tree (t1,t2, Ce_ident split_id1))
		  (spat_one (Csp_tree (t1,t2, Ce_ident split_id2)))))
      else None
  | Csp_indpred(id,el) -> None

let derive_maybe_alloc (pl,sl) = (* pairs (e,f) where e is allocated if e!=f is added *)
  let f sp l = match sp with
    | Csp_node _ -> l
    | Csp_listsegi(SINGLE _,e1,f1,_,_,_)  ->
	if equal_exp e1 f1 || is_uneq pl (e1,f1) then l else (e1,f1)::l
    | Csp_listsegi(k,e1,f1,e2,f2,_) ->
	if equal_exp e1 f1 || equal_exp e2 f2 || is_uneq pl (e1,f1) || is_uneq pl (e2,f2) then l
	else  (e1,f1)::(e2,f2)::l
    | Csp_tree(_,_,e) ->
	if equal_exp (Ce_num 0) e || is_uneq pl (e,Ce_num 0) then l
	else (e,Ce_num 0)::l
    | Csp_indpred _  -> l in
  spat_fold f sl []

let expose_ptsto cp epto =
  let norm_cf ((pl,sl),scpl) = 
    aggr_norm_pure pl >>= fun pl ->
    Some ((pl,sl),scpl) in
  map_option (fun ((pl,sl),scpl) ->
    let epto = (eq_to_sub (unSome (eq_add_list empty_eq pl))) epto in
    try
      let (pl1,sl1),sl2 = split_spatial (unroll_can_sp pl epto) sl in
      Some [((pure_conj pl1 pl,spat_star sl1 sl2),scpl)]
    with Not_found ->
      let alloc = derive_maybe_alloc (pl,sl) in
      try 
        let (_,f) = List.find (fun (e,f) -> equal_exp e epto) alloc in
        Some [((pure_conj_one (mk_EQ(epto,f)) pl,sl),scpl);((pure_conj_one (mk_NEQ(epto,f)) pl,sl),scpl)]
      with Not_found -> None) cp.cfl >>= fun cfll ->
  Some {cfl=map_partial norm_cf (List.concat cfll)}

let remove_empty_preds (pl,sl) = (* remove spatial predicates when they're empty *)
  spat_fold
    (fun s (pl,sl) -> match s with
       | Csp_node _ -> (pl,sl)
       | Csp_listsegi(SINGLE (_,"@list",_),e1,f1,e2,f2,ds) ->
	   let elim1 = equal_exp e1 f1 && dset_mem f1 ds in
	   if not elim1 then (pl,sl)
	   else (pure_conj_one (mk_EQ(e2,mk_fun ("@list",[]))) pl,spat_remove sl s) 
       | Csp_listsegi(SINGLE (_,"@set",_),e1,f1,e2,f2,ds) ->
	   let elim1 = equal_exp e1 f1 && dset_mem f1 ds in
	   if not elim1 then (pl,sl)
	   else (pure_conj_one (mk_EQ(e2,mk_fun ("@set",[]))) pl,spat_remove sl s) 
       | Csp_listsegi(SINGLE _,e1,f1,e2,f2,ds) ->
	   let elim1 = equal_exp e1 f1 && dset_mem f1 ds in
	   if not elim1 then (pl,sl)
	   else (pure_conj_one (mk_EQ(e2,Ce_num 0)) pl,spat_remove sl s)
       | Csp_listsegi(k,e1,f1,e2,f2,ds) ->
	   let elim1 = equal_exp e1 f1 && dset_mem f1 ds in
	   let elim2 = equal_exp e2 f2 && dset_mem e2 ds in
	   if not elim1 && not elim2 then (pl,sl)
	   else (pl,spat_remove sl s)
       | Csp_tree(_,_,e) ->
	   if not (equal_exp e (Ce_num 0)) then (pl,sl)
	   else (pl,spat_remove sl s)
       | Csp_indpred (id,el) -> (pl,sl)) sl (pl,sl)

let generate_ptsto = (* generate constraints Es!=Es' => {Fs} meaning Fs allocated *)
  let rec f = function
    | Csp_node (_,e,_) ->
        [([],[e])]
    | Csp_listsegi(SINGLE _,e1,f1,_,_,_) ->
        [([(e1,f1)],[e1])]
    | Csp_listsegi((DOUBLE _|XOR _),e1,f1,e2,f2,_) ->
        [([(e1,f1)],[e1;f2]);([(e2,f2)],[e1;f2])]
    | Csp_tree(_,_,e) ->
        [([(e,Ce_num 0)],[e])]
    | Csp_indpred (i,el) ->
        let body = instance_ip_lhs (i,el) in
        List.concat (List.map g body)
  and f_inner gl = function
    | Csp_indpred _ -> []
    | sp -> List.map (fun (gl',el) -> (gl@gl',el)) (f sp)
  and g (pl,((_,sl),_)) =
    let pl = pure_to_list pl in
    if List.for_all is_NEQ pl then
      let gl = List.map un_NEQ pl in
      spat_fold (fun sp sl1 -> f_inner gl sp @ sl1) sl []
    else
      []
  in f

let add_constraints sl =
  let constrs = ref empty_constrs in
  let add c = constrs := add_constr !constrs c in
  let add_basic_constrs () =
    let f = function
      | Csp_listsegi(SINGLE _,e1,f1,e2,f2,ds) ->
	  let el = dset_to_list ds in
	  List.iter (fun e -> add ([(e1,f1)],(e,e1))) el
      | Csp_listsegi((DOUBLE _|XOR _),e1,f1,e2,f2,ds) ->
	  let el = dset_to_list ds in
	  List.iter (fun e -> add ([(e1,f1)],(e,e1))) el;
	  add ([(e1,f1)],(e2,f2));
	  List.iter (fun e -> add ([(e1,f1)],(e,e1))) el;
	  List.iter (fun e -> add ([(e1,f1)],(e,f2))) el;
	  add ([(e2,f2)],(e1,f1));
	  List.iter (fun e -> add ([(e2,f2)],(e,e1))) el;
	  List.iter (fun e -> add ([(e2,f2)],(e,f2))) el
      | _ -> ()
    in spat_fold (fun x () -> f x) sl () in
  let ptsto_list =  spat_fold (fun sp r -> generate_ptsto sp @ r) sl [] in
  let add_notnil_constrs () =
    let f (ineql,alloc) = List.iter (fun e -> add (ineql,(e,Ce_num 0))) alloc
    in List.iter f ptsto_list in
  let add_star_constrs () =
    let f ((ineql1,alloc1),(ineql2,alloc2)) =
      let ineql = ineql1@ineql2
      in iter_pairs (fun (e1,e2) -> add (ineql,(e1,e2))) alloc1 alloc2
    in iter_pairwise f ptsto_list
  in add_basic_constrs();
    add_notnil_constrs ();
    add_star_constrs ();
    !constrs
    
(* substitute equalities eq into cf and normalize *)
let apply_equalities_uform eq uf =
  let sub = eq_to_sub eq in
  match map_form sub (uf,[]) with
    | None -> None
    | Some ((pl,sl),_) ->
	Some ((*List.stable_sort compare_atom*) pl, sl)

let apply_equalities_form eq cf =
  let sub = eq_to_sub eq in
  match map_form sub cf with
    | None -> None
    | Some ((pl,sl),scpl) ->
	Some (((*List.stable_sort compare_atom*) pl, sl),scpl)

let apply_equalities_list eq el =
  let sub = eq_to_sub eq in
  List.map sub el

let apply_equalities_prop eq prop =
  let sub = eq_to_sub eq in
  normalize_can_prop (map_prop sub prop)

(* transfer equalities from pl to eq and normalize *)
let update_equalities eq (pl,sl) =
  eq_partition_list eq pl >>= fun (eq,pl) ->
  apply_equalities_uform eq (pl,sl) >>= fun (pl,sl) ->
  Some (eq,(pl,sl))

let (@@) (x1,y1) (x2,y2) = (pure_conj x1 x2,spat_star y1 y2)
let (%) ((pl,sl),scpl) pl' = ((pure_conj pl' pl,sl),scpl)


(* give priority to psto's with existential lhs *)
let compare_can_spred csp1 csp2 =
  let lhs = function
    | Csp_listsegi (_,e,f,_,_,_) -> if existential_id f then (existential_id e, 2) else (existential_id e, 3)
    | Csp_node (_,e,_) -> (existential_id e, 1)
    | Csp_tree (_,_,e) -> (existential_id e, 3)
    | Csp_indpred (_,el) -> (List.exists existential_id el, 3)
  in compare (lhs csp1) (lhs csp2) 

(* unroll inductive predicates in sl using consequences of pl *)
let unroll_ip ~is_lhs pl sl =
  let rec g body = match body with
     [] -> None
   | (pl',(uf,[]))::body ->
	(try
	  if pure_fold (fun ca x -> x && eval_can_atom pl ca) pl' true
	  then Some uf
	  else g body
	with Not_found -> g body)
   | (pl',(uf,_::_))::body -> assert false (* No boxes in inductive predicates *) in
  let rec f sp = match sp with
    | Csp_indpred (i,el) ->
	let body = (if is_lhs then instance_ip_lhs else instance_ip_rhs) (i,el) in
	(match g body with
	  | None -> (pure_true,spat_one sp)
	  | Some uf -> f_uf uf)
    | sp -> (pure_true,spat_one sp)
  and f_uf (pl1,sl1) = 
    spat_fold (fun sp cf -> f sp @@ cf) sl1 (pl1,spat_empty)
  in
  f_uf (pure_true,sl)

let unroll_ip_lhs pl sl =
  unroll_ip ~is_lhs:true pl sl

let unroll_ip_rhs pl sl =
  let (pl1,sl1) = unroll_ip ~is_lhs:false pl sl
  in (pl1, (*List.stable_sort compare_can_spred*) sl1)

let update_disjoint_sets eq cs (pl,sl) = (* update disjoint set in ls's *)
  let imply_uneq pl (e1,e2) = (* checks pl |- e1!=e2 *)
    match eq_add eq (e1,e2) with
      | None -> true
      | Some eq1 -> 
	  match update_equalities eq1 (pl,spat_empty) with
	    | None -> true
	    | Some (eq2,(pl1,_)) ->
		let cs' = sub_constrs (eq_to_sub eq) cs in
		match solve_constraints eq2 pl1 cs' with
		| None -> true | Some _ -> false in
  let sp_alloc_list = spat_fold (fun sp r -> (sp, generate_ptsto sp)::r) sl [] in
  let rec g = function
    | ((sp,cstrl)::l,l') ->
      begin match sp with
      | Csp_listsegi(k,e1,f1,e2,f2,ds) ->
	  let el' = dset_to_list ds in
	  let el'' =
	    let f (eel,el) =
	      List.filter
	        (fun e ->
	          let pl1 = List.fold_left (fun r e1 -> pure_conj_one (mk_NEQ(e,e1)) r)
	    	   (pure_one (mk_NEQ (e,Ce_num 0))) el'
	          in List.for_all (imply_uneq (pure_conj pl pl1)) eel) el
	    in List.flatten(List.map f (List.flatten (List.map snd (l@l'))))
	  in
	  let ds' = dset_from_list (el'@el'') in
	  let sp' = Csp_listsegi(k,e1,f1,e2,f2,ds')
	  in g (l,(sp',cstrl)::l')
      | _ -> g (l,(sp,cstrl)::l')
      end
    | ([],l') -> l'
  in List.fold_left (fun r (x,_) -> spat_star_one x r) spat_empty (g (List.rev sp_alloc_list,[]))

let compare_string_can_prop (s1,p1) (s2,p2) =
  let n = String.compare s1 s2
  in if n<>0 then n
    else compare_prop p1 p2


(** [normal_form sl_frame eq cf] normalizes [cf] 
    assuming that [eq] and [sl_frame] hold separately. *)
let rec normal_form sl_frame eq ((pl,sl),scpl) : (equalities * can_form) option =
  let (pl,sl) =
    if !Config.inductive_preds
    then (pl,spat_empty) @@ unroll_ip_lhs pl sl
    else (pl,sl) in
  update_equalities eq (pl,sl) >>= fun (eq,(pl,sl)) ->
  let cs = add_constraints (spat_star sl_frame sl) in
  (*
    Format.fprintf !Config.formatter "@.@.NORMAL FORM CONSTRAINT START@.";
    pp_constraints !Config.formatter cs;
  *)
  solve_constraints eq pl cs >>= fun (eq,pl,cs) ->
  (*
    Format.fprintf !Config.formatter "@.@.NORMAL FORM CONSTRAINT END@.";
    pp_constraints !Config.formatter cs;
  *)
  apply_equalities_uform eq (pl,sl) >>= fun (pl,sl) ->
  let (pl,sl) =
    if !Config.inductive_preds
    then (pl,spat_empty) @@ unroll_ip_lhs pl sl
    else (pl,sl) in
  let (pl,sl) = remove_empty_preds (pl,sl) in
  update_equalities eq (pl,sl) >>= fun (eq,(pl,sl)) ->
  let pl = remove_atom_duplicates pl in
  let sl = update_disjoint_sets eq cs (pl,sl) in
  let f (eq,scpl) (x,p) =
    let (eq',p') = normal_form_prop (spat_star sl sl_frame) eq (and_cprop_pure p pl) in
    match p'.cfl with [] -> None | _::_ -> Some (eq',(x,p')::scpl) in
  lifted_fold_left f (Some (eq,[])) scpl >>= fun(eq,scpl) ->
  apply_equalities_uform eq (pl,sl) >>= fun cf ->
  let scpl = List.map (fun (x,p) -> (x,apply_equalities_prop eq p)) scpl in
  if List.exists (fun (_,p)-> match p.cfl with [] -> true | _::_ -> false) scpl then
    None
  else
    let scpl = List.stable_sort compare_string_can_prop scpl in
    Some (eq,(cf,scpl))

and normal_form_prop sl_frame eq p : (equalities * can_prop) = 
  let f (eq1,cfl1) (eq2,cf2) = 
    let (eq,eq1',eq2') = combine_eq eq1 eq2 in
    (eq, (cf2 % eq_to_can_pure eq2) 
         :: List.map (fun cf->cf (*% eq_to_can_pure eq1*)) cfl1) in
  let g (eq,cfl) = (eq, {cfl=remove_can_form_duplicates cfl}) in
  let eqcfl = List.rev (map_partial (normal_form sl_frame eq) p.cfl) in
  match eqcfl with
  | [] -> (eq,{cfl=[]})
  | (eq,_)::_ -> g (List.fold_left f (eq,[]) eqcfl)

(** Entry point: normalize can_form *)
let normal_can_form cf =
  normal_form spat_empty empty_eq cf >>= fun (eq,cf) ->
  Some (cf % eq_to_can_pure eq)

let check_atom_entailment pl ca =
  match entails_atom pl ca with
  | Some b -> b
  | None -> false

let (>) cel t =
  try fld_get t cel
  with Not_found -> Ce_ident (gensym_garb (create_ident t))

type ctx = {ctx_eq: equalities;
            ctx_inst: (ident * can_exp) list}

(** substitute out existential instantiations *)
let sub_existentials inst' uf =
  let sub = mk_subst inst' in
  match map_uform sub uf with
   | None -> raise No_frame_found
   | Some uf' -> uf'

(** instantiate existential ex as ex=e in ctx and uf *) 
let inst_sub_existential (ex,e) ctx uf =
  let x = (match ex with | Ce_ident x -> x | _ -> assert false) in
  let sub = mk_single_subst (x,e) in
  let uf' = match map_uform sub uf with 
    | None -> raise No_frame_found
    | Some uf' -> uf' in
  ({ctx with ctx_inst = (x,e)::ctx.ctx_inst}, uf')

(** instantiate xe=e' and substitute in el *)
let inst_sub_existential_elist (xe,e') (inst,el) =
  let x = match xe with | Ce_ident x -> x | _ -> assert false in
  let sub = mk_single_subst (x,e') in
  let el' = List.map (fun (t,e) -> (t,sub e)) el
  in ((x,e')::inst,el')

(** add instantiations for _x=e on the rhs. For each lseg(x,y,S) replace _z in S by x!=_z  *)
let preprocess_existentials eq (cf,inst) =
  let f ca res =
    let rec g res = function
      | (Ce_ident i,e) when is_existential i -> (i,e)::res
      | (e,Ce_ident i) when is_existential i -> (i,e)::res
(*      | (Ce_fun(op1,args1),Ce_fun(op2,args2)) ->
	  if op1=op2 then List.iter2 (fun a b -> g (a,b)) args1 args2 *)
      | _ -> res in
    if is_EQ ca then g res (un_EQ ca) else res in
  let new_inst = pure_fold f (fst cf) [] in
  (match new_inst with [] -> Some cf | _::_ ->
     let pl = List.fold_left (fun r (i,e) -> pure_conj_one (mk_EQ(Ce_ident i,e)) r) pure_true new_inst in
     eq_add_list eq pl >>= fun eq' ->
     apply_equalities_uform eq' cf)
  >>= fun(pl,sl) ->
  let inst' = new_inst @ inst in
(*  let new_neqs = ref [] in
  let f = function
    | Csp_listsegi (SINGLE(snode,sfun),e1,f1,e2,f2,ds) ->
	let (el_ex,ds') = dset_partition (fun e -> existential_id e && e<>e1 && e<>f1) ds in
	let _ = new_neqs := !new_neqs @ List.map (fun e -> mk_NEQ (e1,e)) el_ex
	in Csp_listsegi (SINGLE(snode,sfun),e1,f1,e2,f2,ds')
    | sp -> sp in
  let sl = List.map f sl in *)
  Some ((pl,sl),inst')

 (** default return value when we hit inconsistent on the lhs *)
let ret_false = (can_prop_false,[])

(** finds [pl'] and [sl'] such that  [pl1 * sl1 * pl' |- pl2 * sl2 * sl'] *)
let find_frame_uform (eq,inst,fr) (pl1,sl1) (pl2,sl2) =
  let print_rule s = (if !Config.verbose2 then fprintf !Config.formatter "[RULE %s]@." s) in
  let error () = raise No_frame_found in
  let none_is_error = function Some x -> x | None -> error() in
  let none_do_f f = function Some x -> x | None -> f() in
  let sl2 = List.stable_sort compare_can_spred sl2 in
  let _ =
    if !Config.verbose2 then
      fprintf !Config.formatter "Entailment:@.%a@." pp_can_entailment ({cfl=[((pl1,sl1),[])]},{cfl=[((pl2,sl2),[])]}) in
  let ((pl2,sl2),inst) = none_is_error (preprocess_existentials eq ((pl2,sl2),inst)) in
  let rec f_ifthen (eq',inst) {cfl=cfl} uf2 =  (* case analysis *)
    let go (res,inst) (uf,scpl) = 
      match normal_form fr eq' (uf,[]) with
	| None -> ret_false
	| Some (eq'',(uf,_)) ->
	    let (pl2',sl2') = none_do_f (fun()-> can_uform_false) (apply_equalities_uform eq'' uf2) in
	    let sl2'' = List.stable_sort compare_can_spred sl2' in
	    let (cp1,inst1) = f {ctx_eq=eq''; ctx_inst=inst} uf (pl2',sl2'') in
	    let inst1 = List.filter (fun (x,_) -> not (List.exists (fun (y,_) -> equal_ident x y) inst)) inst1 in
            let new_pl = eq_to_can_pure eq'' in
	    let new_pl = List.fold_left (fun res (x,e)-> pure_conj_one (mk_EQ(Ce_ident x,e)) res) new_pl inst1 in
	    (can_prop_or res (and_cprop_pure cp1 new_pl), inst) in
    (if !Config.verbose2 && List.length cfl <> 1 then fprintf !Config.formatter "[CASE SPLIT %n]@." (List.length cfl));
    List.fold_left go ({cfl=[]},inst) cfl
  and f ctx (pl1,sl1) (pl2,sl2) =  (* main subtraction function *)
    let (pl2,sl2) =
      if !Config.inductive_preds
      then (pl2,[]) @@ unroll_ip_rhs pl1 sl2
      else (pl2,sl2) in
    let (_,sl1) = none_do_f (fun()->assert false) (normalize_can_uform (pl1,sl1)) in
    let (_,sl2) = none_do_f (fun()->assert false) (normalize_can_uform (pl1,sl2)) in
    let sl2 = List.stable_sort compare_can_spred sl2 in
      (if !Config.verbose2 then
	fprintf !Config.formatter "@.Subtraction Entailment:@.%a@.%a@."
	  pp_can_entailment ({cfl = [((pl1,sl1),[])]},{cfl=[((pl2,sl2),[])]}) pp_inst ctx.ctx_inst);
      match sl2 with
	| [] ->
  	    let ((pl2,_),inst) = none_is_error (preprocess_existentials eq ((pl2,[]),ctx.ctx_inst)) in
	    if pure_fold (fun ca b -> b && check_atom_entailment pl1 ca) pl2 true then begin
	      print_rule "axiom";
              ({cfl = [((pure_conj (eq_to_can_pure ctx.ctx_eq) pl1,sl1),[])]},inst)
	    end else error()
	| Csp_node(nid,e,el)::sl' ->
	    (try
	      let (el',sl1b) = split_spatial (node_filter nid e) sl1 in
	      let (inst',pl3) = match fld_subset ctx.ctx_inst el' el with 
		| None -> none_is_error (weak_fld_subset ctx.ctx_inst el' el)
		| Some x -> x in
	      let (pl2',sl'') = none_is_error (map_uform (mk_subst inst') (pure_conj pl3 pl2,sl')) in
	      print_rule "ptsto";
	      f {ctx with ctx_inst = inst'} (pl1,sl1b) (pl2',sl'')
	    with Not_found ->
	      match expose_ptsto {cfl=[((pl1,sl1),[])]} e with
(*		| Some {cfl = [((pl1',sl1'),_)]} -> 
		  if pl1=pl1' then f ctx (pl1',sl1') (pl2,sl2) else 
		  let (eq',(uf1',_)) = none_is_error (normal_form ctx.ctx_eq ((pl1',sl1'),[])) in 
		  let uf2' = none_is_error (apply_equalities_uform eq' (pl2,sl2)) in
		  f {ctx with ctx_eq=eq'} uf1' uf2' *)
		| Some cp -> 
		    print_rule "ptsto_expose";
		    f_ifthen (ctx.ctx_eq,ctx.ctx_inst) cp (pl2,sl2)
		| None -> error())
	| csp::sl' when List.exists (fun csp' -> can_spred_leq ctx.ctx_inst csp' csp <> None) sl1 ->  (* frame for ls's and trees *)
	    print_rule "frame";
	    let ((inst,pl3),sl1b) = split_spatial (fun csp' -> can_spred_leq ctx.ctx_inst csp' csp) sl1 in
	    let (pl2,sl') = none_is_error (map_uform (mk_subst inst) (pure_conj pl3 pl2,sl')) in
	    f {ctx with ctx_inst=inst} (pl1,sl1b) (pl2,sl')
	| Csp_listsegi(SINGLE (_,sfun,_),e1,f1,e2,_,_)::sl' when equal_exp e1 f1 ->
	    print_rule "emp_lseg";
            let (inst',pl_new) = none_is_error (weak_unify_exp ctx.ctx_inst e2 (mk_fun(sfun,[]))) in
	    let (pl2,sl') = none_is_error (map_uform (mk_subst inst') (pure_conj pl_new pl2,sl')) in
            f {ctx with ctx_inst=inst'} (pl1,sl1) (pl2,sl')
	| Csp_listsegi((DOUBLE _|XOR _),e1,f1,e2,f2,_)::sl' when equal_exp e1 f1 && equal_exp e2 f2 ->
	    print_rule "emp_lseg2";
	    f ctx (pl1,sl1) (pl2,sl')
	| Csp_listsegi(SINGLE(snode,sfun,cel0),e1,f1,e2,_,ds)::sl' ->
	    let rec g = function
	      | (sl1a, sp::sl1b) ->
		    begin match sp with
		      | Csp_node(nid,e,cel)
			  when equal_exp e1 e && dset_notmem pl1 e ds ->
			  print_rule "lseg_ptsto";
			  let (e3,e4,nid',cel',pl_new) = unroll_node_rhs snode e in
			  if Pervasives.(=) nid' nid then begin
			    let e5 = Ce_ident (gensym_garb (create_ident "VAL")) in
			    let pl_new = pure_conj (none_is_error (mk_EQ_exn (e2,mk_fun(sfun,[e4;e5])))) pl_new in
			    let cel' = fld_union cel' cel0 in
			    let sl_new = [Csp_node(nid,e,cel');Csp_listsegi (SINGLE(snode,sfun,cel0),e3,f1,e5,Ce_num 0,ds)] in
			    let ((pl_new,sl_new),inst) =
			      none_is_error (preprocess_existentials ctx.ctx_eq ((pl_new,sl_new),ctx.ctx_inst)) in
			    f {ctx with ctx_inst=inst} (pl1,sl1) (pure_conj pl_new pl2,sl_new@sl')
		          end else (* TODO -- List is empty *)
			    assert false
		      | Csp_node(nid,e,cel)
			  when equal_exp e1 e && existential_id f1 && dset_notmem pl1 e ds ->
			  print_rule "lseg_ptsto_inst_A";
			  let (e3,e4,nid',cel',pl_new) = unroll_node_rhs snode e in
			  if Pervasives.(=) nid' nid then begin
			    let e5 = Ce_ident (gensym_garb (create_ident "VAL")) in
			    let pl_new = pure_conj (none_is_error (mk_EQ_exn (e2,mk_fun(sfun,[e4;e5])))) pl_new in
			    let cel' = fld_union cel' cel0 in
			    let sl_new = [Csp_node(nid,e,cel');Csp_listsegi (SINGLE(snode,sfun,cel0),e3,f1,e5,Ce_num 0,ds)] in
			    let ((pl_new,sl_new),inst) =
			      none_is_error (preprocess_existentials ctx.ctx_eq ((pl_new,sl_new),ctx.ctx_inst)) in
			    let (_,inst') = f {ctx with ctx_inst=inst} (pl1,sl1) (pure_conj pl_new pl2,sl_new@sl') in
			    let uf2' = sub_existentials inst' (pl2,sl2) in
			    print_rule "lseg_ptsto_inst_B";
			    f ctx (pl1,sl1) uf2'
		          end else (* TODO -- List is empty *)
			    assert false
		      | Csp_listsegi(SINGLE(snode',sfun',cel0'),e1',f1',e2',_,ds')
                          when (Pervasives.(=) sfun "@zero" || Pervasives.(=) sfun sfun') 
				&& Pervasives.(=) snode' snode && equal_exp e1' e1 
				&& not(existential_id e1) && existential_id f1 && dset_subset ds ds'->
			  print_rule "lseg_lseg_A";
			  let (inst',pl3) = none_is_error (fld_subset ctx.ctx_inst cel0' cel0) in
			  assert (Pervasives.(=) inst' ctx.ctx_inst && pure_is_true pl3); (* No existential vars must exist in cel0 or cel0' *)
			  let e5 = Ce_ident (gensym_garb (create_ident "VAL")) in
			  let (_,inst') = f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi(SINGLE(snode,sfun',cel0),f1',f1,e5,Ce_num 0,ds)::sl')) in
			  let uf2' = sub_existentials inst' (pl2,sl2) in
			  print_rule "lseg_lseg_B";
			  f {ctx with ctx_inst=inst'} (pl1,sl1) uf2'  
		      | Csp_listsegi(SINGLE(snode',sfun',cel0'),e1',f1',e2',_,ds')
                          when (Pervasives.(=) sfun "@zero" || Pervasives.(=) sfun sfun') 
				&& Pervasives.(=) snode' snode && equal_exp e1' e1 
				&& not(existential_id e1) && existential_id f1 ->
			  print_rule "lseg_lseg2";
			  let (ctx',uf2') = inst_sub_existential (f1,f1') ctx (pl2,sl2) in
			  f ctx' (pl1,sl1) uf2'
		      | Csp_listsegi(SINGLE(snode',sfun',cel0'),e1',f1',e2',_,ds')
                          when (Pervasives.(=) sfun "@zero" || Pervasives.(=) sfun sfun') 
				&& Pervasives.(=) snode' snode && equal_exp e1' e1 
	                        && dset_subset ds ds' ->
			  print_rule "lseg_lseg3";
			  using_induction ();
			  let (inst',pl3) = none_is_error (fld_subset ctx.ctx_inst cel0' cel0) in
			  assert (Pervasives.(=) inst' ctx.ctx_inst && pure_is_true pl3); (* No existential vars must exist in cel0 or cel0' *)
			  let e5 = Ce_ident (gensym_garb (create_ident "VAL")) in
			  let pl_new = pure_conj (none_is_error (mk_EQ_exn (e2,mk_fun(sfun,[e2';e5])))) pl2 in
			  let sl_new = (Csp_listsegi(SINGLE(snode,sfun',cel0),f1',f1,e5,Ce_num 0,ds))::sl' in
			  let ((pl_new,sl_new),inst) =
			    none_is_error (preprocess_existentials ctx.ctx_eq ((pl_new,sl_new),ctx.ctx_inst)) in
			  f {ctx with ctx_inst=inst} (pl1,List.rev_append sl1a sl1b) (pl_new,sl_new)
		      | _ -> g (sp::sl1a,sl1b)
		    end
	      | (sl1a,[]) -> error()
	    in
	      if not (existential_id e1) && existential_id f1
	      then 
		try  g ([],sl1)
		with No_frame_found ->
		  print_rule "lseg_inst_guess";
		  let (ctx',uf2') = inst_sub_existential (f1,e1) ctx (pl2,sl2) in
                  f ctx' (pl1,sl1) uf2'
	      else g ([],sl1)
	| Csp_listsegi(DOUBLE(t1,t2),e1,f1,e2,f2,ds)::sl' ->
	    let rec g = function
	      | (sl1a, sp::sl1b) ->
		    begin match sp with
		      | Csp_node("Node",e,cel)
			  when equal_exp e1 e && fld_mem (t2,e2) cel && is_uneq pl1 (e2,f2) && dset_notmem pl1 e ds ->
			  let e3 = cel>t1
			  in f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi (DOUBLE(t1,t2),e3,f1,e1,f2,ds)::sl'))
		      | Csp_node("Node",e,cel)  (* |-> on the rhs *)
			  when equal_exp f2 e && fld_mem (t1,f1) cel && dset_notmem pl1 e ds ->
			  using_induction ();
			    let e3 = cel>t2
			    in f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi (DOUBLE(t1,t2),e1,f2,e2,e3,ds)::sl'))
		      | Csp_listsegi(DOUBLE(t1',t2'),e1',f1',e2',f2',ds')
			  when Pervasives.(=) t1 t1' && Pervasives.(=) t2 t2' 
			     && equal_exp e1 e1' && equal_exp e2 e2' && dset_subset ds ds' ->
			  using_induction ();
			    f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi(DOUBLE(t1,t2),f1',f1,f2',f2,ds)::sl'))
		      | _ -> g (sp::sl1a,sl1b)
		    end
	      | (sl1a,[]) -> error()
	    in
	      g ([],sl1)
	| Csp_listsegi(XOR t1,e1,f1,e2,f2,ds)::sl' ->
	    let rec g = function
	      | (sl1a, sp::sl1b) ->
		    begin match sp with
		      | Csp_node("Node",e,cel)
			  when equal_exp e1 e && is_uneq pl1 (e2,f2) && dset_notmem pl1 e ds ->
			  let e3 = cel>t1
			  in f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi (XOR t1,mk_XOR(e3,e2),f1,e1,f2,ds)::sl'))
		      | Csp_node("Node",e,cel)  (* |-> on the rhs *)
			  when equal_exp f2 e && dset_notmem pl1 e ds ->
			  using_induction ();
			    let e3 = cel>t1
			    in f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi (XOR t1,e1,f2,e2,mk_XOR(e3,f1),ds)::sl'))
		      | Csp_listsegi(XOR t1',e1',f1',e2',f2',ds')
			  when Pervasives.(=) t1 t1' 
			     && equal_exp e1 e1' && equal_exp e2 e2' && dset_subset ds ds' ->
			  using_induction ();
			    f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi(XOR t1,f1',f1,f2',f2,ds)::sl'))
		      | Csp_listsegi(XOR t1',f2',e2',f1',e1',ds') (* or the wrong way round *)
			  when Pervasives.(=) t1 t1' 
			     && equal_exp e1 e1' && equal_exp e2 e2' && dset_subset ds ds' ->
			  using_induction ();
			    f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_listsegi(XOR t1,f1',f1,f2',f2,ds)::sl'))
		      | _ -> g (sp::sl1a,sl1b)
		    end
	      | (sl1a,[]) -> error()
	    in
	      g ([],sl1)
	| Csp_indpred (id,el)::sl' -> error ()
	| Csp_tree(_,_,Ce_num 0)::sl' -> f ctx (pl1,sl1) (pl2,sl')
	| Csp_tree(t1,t2,e)::sl' -> 
	    let rec g = function
	      | (sl1a, sp::sl1b) ->
		    begin match sp with
		      | Csp_node("Node",e1,cel)
			  when equal_exp e e1 &&
			    fld_exists (fun (t',_) -> Pervasives.(=) t' t1) cel &&
			    fld_exists (fun (t',_) -> Pervasives.(=) t' t2) cel ->
			  let e2 = fld_get t1 cel in
			  let e3 = fld_get t2 cel in
			  f ctx (pl1,List.rev_append sl1a sl1b) (pl2,(Csp_tree(t1,t2,e2)::Csp_tree(t1,t2,e3)::sl'))
		      | Csp_tree(t1',t2',e') 
			  when Pervasives.(=) t1 t1' && Pervasives.(=) t2 t2' && equal_exp e e' ->
			  f ctx (pl1,List.rev_append sl1a sl1b) (pl2,sl')
		      | _ -> g (sp::sl1a,sl1b)
		    end
	      | (sl1a,[]) -> error()
	    in g ([],sl1) in
  let ctx = {ctx_eq=eq; ctx_inst=inst} in
  let (res,inst) = f ctx (pl1,sl1) (pl2,sl2) in
  let new_pl = List.fold_left (fun res (x,e)-> pure_conj_one (mk_EQ(Ce_ident x,e)) res) pure_true inst in
  ({cfl = List.map (fun ((pl,sl),scpl) -> ((pure_conj new_pl pl,sl),scpl)) res.cfl}, inst)

let reduce_cfl (eq,pl) cfl2 = 
  let rec is_valid pll = 
    let pll = map_partial aggr_prove_pure pll in
(*    let pll = remove_atom_list_duplicates pll in *)
    let pll = List.sort (fun x y-> pure_length x - pure_length y) pll in
(*    (if !Config.verbose2 && pll<>[[]] then fprintf !Config.formatter "reduce_cfl: %a@." 
      pp_can_prop {cfl=List.map (fun pl->((pl,[]),[])) pll}); *)
    match pll with
      | [] -> false
      | pl::pll -> 
    match pure_to_list pl with
      | [] -> true
      | [ca] -> 
          let pll = List.filter (fun pl -> not (mem_atom ca pl)) pll in
	  let nca = neg_can_atom ca in
	  let pll = List.map (pure_filter (fun x->x<>nca)) pll in 
	  is_valid pll 
      | (ca::_::_) -> (* do a case split *)
	  let nca = neg_can_atom ca in
          let pll1 = List.filter (fun pl -> not (mem_atom ca pl)) pll in
	  let pll1 = List.map (pure_filter (fun x->x<>nca)) pll1 in 
          let pll2 = List.filter (fun pl -> not (mem_atom nca pl)) (pl::pll) in
	  let pll2 = List.map (pure_filter (fun x->x<>ca)) pll2 in 
	  is_valid pll1 && is_valid pll2 in
	 (*false in*)
  let cfl2 = map_partial (apply_equalities_form eq) cfl2 in
  let cflm2 = map_partial
    (fun (((pl2,_),_)as cf2) -> 
      entails_pure_list pl pl2 >>= fun missing -> 
      (if !Config.verbose2 && not(pure_is_true missing) then
	fprintf !Config.formatter "missing: %a@." 
	  pp_can_form ((missing,[]),[]));
      Some (cf2,missing)) cfl2 in
  if is_valid (List.map snd cflm2) then
    cflm2 
  else 
    List.map (fun (cf,_) -> (cf,pure_true)) cflm2

(** [fold_lhs_can_prop f cp] converts a check cp |- ... into a number 
    of checks cd1 |- ...  cfn |- ... *)
let fold_lhs_can_prop f cp =
  let g e cf = (f cf).cfl @ e
  in {cfl = remove_can_form_duplicates (List.fold_left g [] cp.cfl)}

(** find_frame_form (eq,inst,fr) cf1 cf2 returns (res,inst') such that 
    cf1 |- cf2 * res *)
let rec find_frame_form (eq,inst,fr) cf1 cf2 =
  let with_ok x f = match x with Some y -> f y | None -> {cfl=[]} in
  let with_error x = match x with Some y -> y | None -> raise No_frame_found in
  with_ok (normal_form fr eq cf1) 
  (fun (eq,(uf1,scpl1)) ->
    let (uf2,scpl2) = with_error (apply_equalities_form eq cf2) in
    let (res,inst) = find_frame_uform (eq,inst,fr) uf1 uf2 in
    let (scpl3,scpl4) = List.partition (fun (s,_) -> List.mem_assoc s scpl2) scpl1 in
    let cpcpl = List.map (fun (s,cp) -> (and_cprop_pure cp (fst uf1),List.assoc s scpl2)) scpl3 in
    if List.exists (fun (s,_)-> not (List.mem_assoc s scpl1)) scpl2 then
      raise No_frame_found
    else if List.for_all (fun (cp1,cp2)->entails_prop (eq,inst,(snd uf1)@fr) cp1 cp2) cpcpl then 
      can_prop_star res {cfl = [((pure_true,[]),scpl4)]}
    else 
      raise No_frame_found)


(** [entails_form_prop (eq,inst,fr) cf1 cp2] returns true if [cf1 * fr |- cp2 * fr] *)
and entails_form_prop (eq,inst,fr) cf1 cp2 =
  (if !Config.all_states then fprintf !Config.formatter "Entails_form_prop: %a@ |- %a@."
    pp_can_prop {cfl=[(fun ((pl,sl),scpl) -> ((remove_atom_duplicates pl,sl),scpl)) cf1]} pp_can_prop cp2);
  let g cf1 (cf2,pl) =
    try
      let res = find_frame_form (eq,inst,fr) (cf1 % pl) cf2 in
      !Config.allow_leaks || is_can_prop_empty res
    with No_frame_found -> false in
  match normal_form fr eq cf1 with
  | None -> true
  | Some (eq,cf1) ->
      let ((pl1,_),_) = cf1 in
      let cflm2 = reduce_cfl (eq,pl1) cp2.cfl in
      List.exists (g cf1) cflm2

(** [entails_prop (eq,inst,fr) cp1 cp2] returns true if [cp1 * fr |- cp2 * fr] *)
and entails_prop (eq,inst,fr) cp1 cp2 =
  List.for_all (fun cf1 -> entails_form_prop (eq,inst,fr) cf1 cp2) cp1.cfl

(** find the pure part of a proposition *)
let pure_part cp = 
  let intersection pl pl' = pure_filter (fun ca -> mem_atom ca pl) pl' in
  let rec go = function
    | [] -> pure_true
    | [((pl,_),_)] -> pl
    | ((pl,_),_)::cfl -> intersection pl (go cfl) in
  remove_atom_duplicates (go cp.cfl)


(** [find_frame_form_prop fr cf1 cp2] returns [res] such that [cf1 * fr |- cp2 * fr * res]
    or raises No_frame_found *)
let find_frame_form_prop fr cf1 cp2 =
  (*(if !Config.all_states then fprintf !Config.formatter "find_frame_form_prop: %a@ |- %a@."
    pp_can_prop {cfl=[cf1]} pp_can_prop cp2);*)
  match normal_form fr empty_eq cf1 with
  | None -> {cfl=[]}
  | Some (eq,cf1) ->
      let ((pl1,_),_) = cf1 in
      let cflm2 = reduce_cfl (eq,pl1) cp2.cfl in
      let go (cf2,pl) =
        try Some (find_frame_form (eq,[],fr) (cf1 % pl) cf2)
        with No_frame_found -> None in
      match map_partial go cflm2 with
      | [] -> raise No_frame_found
      | x -> {cfl= List.concat (List.map (fun cp->cp.cfl) x) }

(** [find_frame_prop fr cp1 cp2] returns [res] such that [cp1 * fr |- cp2 * fr * res]
    or raises No_frame_found *)
let find_frame_prop fr cp1 cp2 =
  let (sub_unex,sub_toex) = remove_reinstate_existentials_sub cp1 in
  let cp1' = map_prop sub_unex cp1 in
  let res = fold_lhs_can_prop (fun cf1 -> find_frame_form_prop fr cf1 cp2) cp1' in
  map_prop sub_toex res

(** Does [cf1 |- cf2] ? *)
let entails_form cf1 cf2 =
  try
    let res = find_frame_form (empty_eq,[],[]) cf1 cf2 in
    !Config.allow_leaks || is_can_prop_empty res
  with No_frame_found -> false

let expose_ptsto2 cp e =
  let (eq,cp) = normal_form_prop [] empty_eq cp in
  let cp = and_cprop_pure cp (eq_to_can_pure eq) in
  expose_ptsto cp e

let default_prover =
  { find_ptsto = (fun s e -> split_spatial (node_filter s e));
    expose_ptsto = expose_ptsto2;
    normal_form_prop = normal_form_prop;
    normal_can_form = normal_can_form;
    entails_form = entails_form;
    find_frame_prop = find_frame_prop
  }

