open Misc
open Exp
open Symbtree
open Inductive_preds
open Format

(** total number of iterations for all the calls to mk_stable_single *)
let stabilisation_iterations = ref 0

(* symbolic execution *)
exception Symb_exe_noframe of Symbtree.can_prop * Symbtree.can_prop * string * Location.t
exception Symb_exe_leakmem of Symbtree.can_prop * Location.t
exception Symb_exe_heap_error

let print loc =
  let fmt = !Config.formatter in
  Location.print fmt loc;
  Format.fprintf fmt

(* Option monad *)
let (>>=) x f = 
  match x with 
  | None -> None
  | Some v -> f v

let un_Some xo =
  match xo with
  | None -> assert false
  | Some x -> x

let with_default y x = 
  match x with 
  | None -> y
  | Some v -> v

exception MonadInternal

let lifted_fold_left f e xl =
  let rec go e xl = match (e,xl) with
    | (None,_) -> None
    | (e,[]) -> e
    | (Some e,x::xl) -> go (f e x) xl in
  go e xl

let lifted_map f xl =
  let rec go = function
    | [] -> []
    | x::xl -> match f x  with
               | None -> raise MonadInternal
               | Some y -> y :: go xl
  in try
    Some (go xl)
  with MonadInternal -> None 

let (%) ((pl,sl),scpl) pl' = ((pure_conj pl' pl,sl),scpl)

let pp_frame_inferred (loc,cp) =
  if !Config.verbose1 then
    begin
      if cp <> can_prop_empty
      then print loc "FRAME inferred:@.@[<hov 2>%a@]@." pp_can_prop cp
    end

(** find the pure part of a proposition *)
let pure_part cp = 
  let intersection pl pl' = pure_filter (fun ca -> mem_atom ca pl) pl' in
  let rec go = function
    | [] -> pure_true
    | [((pl,_),_)] -> pl
    | ((pl,_),_)::cfl -> intersection pl (go cfl) in
  remove_atom_duplicates (go cp.cfl)

(** calculate post-condition outside block *)
let exit_from_block prover (rid,modif_body,cq0,loc) cq =
  let rec remove_box rid cp = {cfl = List.map (fun(uf,scpl)->(uf, List.remove_assoc rid scpl)) cp.cfl} in
  let get_shared cq_shared cf =
    let fr = try List.assoc rid (snd cf) with Not_found -> assert false in
    can_prop_or cq_shared (and_cprop_pure (can_prop_star fr cq0) (fst (fst cf))) in
  let cq_local = (* local part of post-condition *) 
    try prover.find_frame_prop [] cq cq0
    with No_frame_found -> raise (Symb_exe_noframe (cq,cq0,"action exit",loc)) in
  let cq_shared = List.fold_left get_shared {cfl=[]} cq_local.cfl in
  let cq_shared = and_cprop_pure cq_shared (pure_part cq_local) in
  let sub = mk_gensym_garb_subst_list (IdSet.elements modif_body) in
  let cq_local = map_prop sub cq_local in
  let cq_local = remove_box rid cq_local in
  let cq_shared = 
    let (eq,p) = prover.normal_form_prop [] empty_eq cq_shared in
    and_cprop_pure p (eq_to_can_pure eq) in
  let sub = mk_gensym_subst_list (IdSet.elements (ex_fv_can_prop cq_local)) in
  let cq_local = map_prop sub cq_local in
  let cq_shared = map_prop sub cq_shared in
  can_prop_star cq_local {cfl=[((pure_true,[]),[(rid,cq_shared)])]}


(** [interfere_once prover act cp] calculate the effect of applying
    interference [act] to [cp]. *)
let interfere_once prover (name,cp_ctx,cp1,cp2) cp =
  (* 1. Rename existential variables in cp_ctx,cp1,cp2 as they might appear in ufl_new *)
  let (cp_ctx,cp1,cp2) = 
    let exfv = IdSet.union (ex_fv_can_prop cp1) 
		(IdSet.union (ex_fv_can_prop cp2) (ex_fv_can_prop cp_ctx)) in
    let sub = existentials_rename_sub exfv in
    (map_prop sub cp_ctx,map_prop sub cp1, map_prop sub cp2) in
  (if !Config.all_states then
    Format.fprintf (!Config.formatter) "go %s (%a,@ %a,@ %a)@ new: %a@." name 
	pp_can_prop cp_ctx pp_can_prop cp1 pp_can_prop cp2 
      pp_can_prop cp);
  (* 2. Compute [cp' = (cp1 -*E cp)]  *)
  let cp' = can_prop_wande cp1 cp in
  (* 3. Remove disjuncts where [cp_ctx -*E cp'] is not satisfiable *)
  let cp' =
    let is_satisfiable cp = List.exists (fun x -> prover.normal_can_form x <> None) cp.cfl in
    {cfl=List.filter (fun cf -> 
        (if !Config.verbose2 then Format.fprintf (!Config.formatter) "___ %a@." pp_can_form cf);
	try is_satisfiable (can_prop_wande cp_ctx {cfl=[cf]})
	with Undef_wand_elim _ -> true) cp'.cfl} in
  (* 3. Compute [cp' * cp2] *)
  let res = can_prop_star cp' cp2 in
  (* (if !Config.all_states then
    Format.fprintf (!Config.formatter) "res0: %a@." pp_can_prop res); *)
  res


(** [make_stable cp rely] returns a [cp'] such that [cp |- cp'] and [stable_under cp rely] *)
let make_stable prover abstraction s cp actl = 
  let uform_list_to_can_prop ufl =
    {cfl = List.map (fun uf -> (uf,[])) ufl} in
  (*
    In the following code, the assertions [ufl_total] and [ufl] are logically equivalent;
    but [ufl_total] will typically contain redundant disjuncts, whereas [ufl] should not.
    The distinction between these two is to work round the possible non-transitivity
    of the prover's entailment checking. 
  *)
  
  (** Fix-point calculation for stability under a single action *)
  let rec mk_stable_single act (ufl_total,ufl,ufl_new) =
    if ufl_new = [] then (ufl_total,ufl) else begin
      incr stabilisation_iterations;
      (* 1. Calculate effect of a single interference step *)
      let ufl' =
        let cp' = interfere_once prover act (uform_list_to_can_prop ufl_new) in
        List.map (fun (uf,scpl) -> assert(scpl=[]); uf) cp'.cfl in
      (* 2. Join *)
      let (ufl_total,ufl,ufl_new) = abstraction.uform_join ufl_total ufl ufl' in
      (if !Config.all_states then Format.fprintf (!Config.formatter)
         "res: %a@." pp_can_prop (uform_list_to_can_prop ufl_new));
      (* 3. Repeat until fixed point *)
      mk_stable_single act (ufl_total,ufl,ufl_new)
    end in
 
  (** Fix-point calculation for stability under a list of actions *)
  let rec mk_stable n (ufl_total,ufl,ufl_new) =
    if ufl_new = [] then ufl else begin
      (if !Config.all_states then
         Format.fprintf (!Config.formatter) "go2(%d,%d,%d)@ old: %a@ new: %a@." 
            n (List.length ufl_total) (List.length ufl_new)
            pp_can_prop (uform_list_to_can_prop ufl_total)
            pp_can_prop (uform_list_to_can_prop ufl_new));
      (* 1. Stabilize with respect to each action separately *)
      let (ufl_total',ufl_new) = 
         let f (t,n) act = mk_stable_single act (t,n,ufl_new) in
         List.fold_left f (ufl_total,[]) actl in
      (* 2. Join *)
      let (_,ufl,ufl_new) = abstraction.uform_join ufl_total ufl ufl_new in
      (* 3. Repeat until fixed point *)
      mk_stable (n+1) (ufl_total',ufl,ufl_new)
    end in

  (if !Config.all_states then Format.fprintf (!Config.formatter) "make_stable@ %a@." pp_can_prop cp);
  let cp = aggr_kill_garbage_vars cp in
  (if !Config.all_states then Format.fprintf (!Config.formatter) "mak2_stable@ %a@." pp_can_prop cp);
  (* pre-processing: substitute out existing existential variables *)
  let cp = 
    let exfv = ex_fv_can_prop_fromTID cp in
    let sub = existentials_rename_tonormal_sub exfv in
    map_prop sub cp in
  (* pre-processing: turn useless variables into existentials *)
  let cp = kill_dead_vars s cp in
  (* main part *)
  let ufl = List.map fst cp.cfl in
  try
    (* 1. Do an initial abstraction *)
    let (ufl_total,ufl,_) = abstraction.uform_join [] [] ufl in
    (* 2. Calculate fix-point *)
    let ufl = mk_stable 0 (ufl_total,ufl,ufl) in
    uform_list_to_can_prop ufl
  with Junk -> can_prop_spred [Csp_node("Junk",Ce_ident (Misc.create_ident "JUNK"),fld_from_list [])]



(** computes the fixpoint of a transformation from propositions to
    propositions, such as the symbolic execution function *)
let compute_transf_fixpoint abstraction (transf: can_prop -> can_prop) cp =
  let rec fix (cp_total,cp,cp_new) =
    if cp_new.cfl = [] then cp
    else fix (abstraction.prop_join cp_total cp (transf cp))
  in
  try
    (* 1. Do an initial abstraction *)
      (* This step is optional: abstracts the initial value.
         Intuitively it should speed up the analysis slightly.
         In practice, it does not affect performance in the 3 list algorithms. *)
    let cp = abstraction.prop_abs cp in
    (* 2. Calculate fix-point *)
    let pfix = fix (cp, cp, cp) in
    let _ = Format.fprintf !Config.formatter "@.FOUND LOOP INVARIANT:@.%a@." pp_can_prop pfix in
    pfix
  with
    Junk -> can_prop_spred [Csp_node("Junk",Ce_ident (Misc.create_ident "JUNK"),fld_from_list [])]






(** Compute [cq] such that [cp |- cq] and [stable_under cp rely] *)
let stabilize prover abstraction (rid,modif,rely) lv cp =
  let get_shared rid (uf,scpl) = try
      ((uf,List.remove_assoc rid scpl), List.assoc rid scpl)
    with Not_found -> assert false in
  let _sub = mk_gensym_garb_subst_list (IdSet.elements modif) in
  let cp = map_prop (existentials_rename_sub (ex_fv_can_prop cp)) cp in
  let cq = (* post-condition outside block *)
    {cfl=List.map (fun cf ->
      let (cf_local,cp_shared) = get_shared rid cf in
      (*let cf_local = unSome (map_form sub cf_local) in *)
      let (sub_unex,sub_toex) = remove_reinstate_existentials_sub {cfl=[cf]} in
      let cp = map_prop sub_unex {cfl=[cf_local]} in
      let cp_2 = map_prop sub_unex {cfl=[((pure_true,snd(fst cf_local)),[])]} in
      let cp_shared = 
        try map_prop sub_toex (prover.find_frame_prop [] (can_prop_star cp cp_shared) cp_2)
        with No_frame_found -> 
	  fprintf !Config.formatter "cp_shared: %a@." pp_can_prop cp_shared;
	  fprintf !Config.formatter "cf_local: %a@." pp_can_prop {cfl=[cf_local]};
          assert false in
      let cp_shared = make_stable prover abstraction lv cp_shared rely in
      (fst cf_local,(rid,cp_shared)::snd cf_local)) cp.cfl}
  in
  (if !Config.all_states then fprintf !Config.formatter "Stabilisation result:@ %a@." pp_can_prop cq);
  cq

(** Result of symbolic execution *)
type se_result = 
  { se_normal : can_prop;
    se_break : can_prop;
    se_cont : can_prop;
    se_return : can_prop }

let se_false = 
  { se_normal = can_prop_false; 
    se_break = can_prop_false;
    se_cont = can_prop_false;
    se_return = can_prop_false }

let se_or se1 se2 = 
  { se_normal = can_prop_or se1.se_normal se2.se_normal; 
    se_break = can_prop_or se1.se_break se2.se_break;
    se_cont = can_prop_or se1.se_cont se2.se_cont;
    se_return = can_prop_or se1.se_return se2.se_return }

let se_map f se = 
  { se_normal = f se.se_normal; 
    se_break  = f se.se_break;
    se_cont   = f se.se_cont;
    se_return = f se.se_return }

type lv_data = 
  { lv_post : IdSet.t;
    lv_break : IdSet.t;
    lv_cont : IdSet.t;
    lv_return : IdSet.t }


let kill_garbage prover lv cp1 =
  let (eq,cp1) = prover.normal_form_prop [] empty_eq cp1 in
  let cp1 = kill_dead_vars lv cp1 in
  let (eq,cp1) = kill_garbage_vars (eq,cp1) in
  let cp1 = and_cprop_pure cp1 (eq_to_can_pure eq) in
  cp1

let calculate_post prover abstraction ht renv (cp1,c,fv2,loc) =

let rec symb_execute (rctx,se1,c,lvs,loc) : se_result =
  let cp1 = 
    let lv = match c with [] -> lvs.lv_post | {can_st_lv = lv}::_ -> lv in
    (if !Config.all_states then fprintf !Config.formatter "lv=%a@." pp_idset lv); 
    kill_garbage prover lv se1.se_normal in
  let _ =
    if !Config.all_states then
      let cloc = match c with {can_st_loc = loc1}::_ -> loc1 | _ -> loc in
      let pp_comp s f cp = if cp.cfl=[] then () else Format.fprintf f "  %s: %a@." s pp_can_prop cp in
      print cloc "Intermediate State:@.%a%a%a[@[%a@]]@.%a@."
        (pp_comp "return") se1.se_return (pp_comp "break") se1.se_break
        (pp_comp "continue") se1.se_cont pp_can_prop cp1 pp_can_cmd c in
  match cp1.cfl with
  | [] -> {se1 with se_normal=can_prop_false}
  | _ -> symb_exe (rctx,{se1 with se_normal = cp1},c,lvs,loc)

and symb_exe (rctx,spre,c,lvs,loc) : se_result =
  let heap_error loc' pp =
    print loc' "ERROR: %a in@ @[%a@]@." pp () pp_can_prop spre.se_normal;
    raise Symb_exe_heap_error in
  match c with
      | [] -> spre
      | {can_st_desc=Cst_nondet (c1,c2)}::c3 ->
	  let lvs' = match c3 with [] -> lvs | {can_st_lv = lv}::_ -> {lvs with lv_post = lv} in
	  let sq1 = symb_execute (rctx,spre,c1,lvs',loc) in
	  let sq2 = symb_execute (rctx,spre,c2,lvs',loc) in
          let sq = se_or sq1 sq2 in
	  symb_execute (rctx,sq,c3,lvs,loc)
      | {can_st_desc=Cst_assign (x,e)}::c1 ->
	  let sub = mk_gensym_garb_subst x in
	  let e' = sub e in
	  let cq = and_cprop_pure (map_prop sub spre.se_normal) (pure_one (mk_EQ (Ce_ident x, e'))) in
	  symb_execute (rctx,{spre with se_normal = cq},c1,lvs,loc)
      | {can_st_desc=Cst_fldlookup (x,e,t);can_st_loc=loc'}::c1 ->
	  let lookup t cel =
            try (cel, fld_get t cel)
            with Not_found ->
              let dummye = Ce_ident (gensym_garb (create_ident "dummy"))
              in (fld_union (fld_from_list [(t,dummye)]) cel, dummye) in
	  (* Old value of x is now garbage *)
	  let sub = mk_gensym_garb_subst x in
	  (* Read from cf *)
          let rec go_simp e rctx ((pl2,sl2),scpl2) =
            let e = (eq_to_sub (unSome (eq_add_list empty_eq pl2))) e in
	    try
	      let (cel,sl2b) = prover.find_ptsto "Node" e sl2 in
	      let (cel',f) = lookup t cel in
	      let f' = sub f in
	      map_form sub ((pl2,Csp_node("Node",e,cel')::sl2b),scpl2) >>= fun cf ->
	      Some {cfl=[cf % pure_one (mk_EQ (Ce_ident x,f'))]}
            with Not_found -> 
	      go_ctx e rctx ((pl2,sl2),scpl2)
	  and go e rctx ((pl2,sl2),scpl2) =
	    (*let _ = if !Config.all_states then fprintf !Config.formatter "go: %d@." (List.length rctx) in*)
            let e = (eq_to_sub (unSome (eq_add_list empty_eq pl2))) e in
	    match prover.expose_ptsto {cfl=[((pl2,sl2),scpl2)]} e with
	      | None -> go_ctx e rctx ((pl2,sl2),scpl2)
	      | Some {cfl=[cf]} -> go_simp e rctx cf
	      | Some (cp) ->
		  let (_,cp) = prover.normal_form_prop [] empty_eq cp in (* needed for termination! *)
		  go_list (go e rctx) cp
	  and go_list f cp =
	    let g res cf = match res with
	      | None -> None
	      | Some res -> match f cf with
		  | None -> None
		  | Some cq -> Some {cfl=cq.cfl@res.cfl} in
	    List.fold_left g (Some {cfl=[]}) cp.cfl
	  and go_ctx e rctx ((pl2,sl2),scpl2) =
	    (*let _ = if !Config.all_states then fprintf !Config.formatter "go_ctx: %d@." (List.length rctx) in*)
            match rctx with
	    | [] -> None
	    | rid::rctx ->
		try match go_list (go e []) (and_cprop_pure (List.assoc rid scpl2) pl2) with
		  | None -> go_ctx e rctx ((pl2,sl2),scpl2)
		  | Some cp -> 
		     let (uf2,scpl2) = unSome (map_form sub ((pl2,sl2),List.remove_assoc rid scpl2)) in
		     Some {cfl=[(uf2,(rid,cp)::scpl2)]}
		with Not_found -> go_ctx e rctx ((pl2,sl2),scpl2) in
	  let cq = match go_list (go e rctx) spre.se_normal with
	    | Some cq-> cq
	    | None -> heap_error loc' (fun fmt () -> fprintf fmt "lookup %a->%s" pp_can_exp e t) in
	  symb_execute (rctx,{spre with se_normal = cq},c1,lvs,loc)
      | {can_st_desc=Cst_fldassign (e1,t,e2);can_st_loc=loc'}::c1 ->
          let go ((pl2,sl2),scpl2) =
            let e1 = (eq_to_sub (unSome (eq_add_list empty_eq pl2))) e1 in
	    let (cel,sl2b) =
	      try prover.find_ptsto "Node" e1 sl2
	      with Not_found -> heap_error loc' (fun fmt () -> fprintf fmt "assign %a->%s" pp_can_exp e1 t) in
	    let cel' = fld_set (t,e2) cel in
	    ((pl2,Csp_node("Node",e1,cel')::sl2b),scpl2) in
	  let cq = match prover.expose_ptsto spre.se_normal e1 with
	    | Some cp -> {cfl=List.map go cp.cfl}
	    | None -> heap_error loc' (fun fmt () -> fprintf fmt "assign %a->%s" pp_can_exp e1 t) in
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_new(x)}::c1 ->
	  let cpre = spre.se_normal in
	  let sub = mk_gensym_garb_subst x in
	  let cp = map_prop sub cpre in
	  let cq = can_prop_star cp (can_prop_spred [Csp_node("Node",Ce_ident x,fld_from_list [])]) in
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_dispose(e);can_st_loc=loc1}::c1 ->
	  let cq =
	    let cp1 = can_prop_spred [Csp_node("Node",e,fld_from_list [])] in
	    try prover.find_frame_prop [] spre.se_normal cp1
	    with No_frame_found -> raise (Symb_exe_noframe (spre.se_normal,cp1,"dispose",loc1)) in
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_pfcall par_calls; can_st_loc=loc1}::c1 ->
(* TODO -- freshen exist. vars *)
	  (* 1. Calculate the specification of a generic command *)
          let go_fun (resid,s,idl,el) =
	    let fi = Hashtbl.find ht s in
	    let idl = resid::idl in
	    let sl_ref = List.combine (create_ident "Result" :: fst fi.fun_param) (List.map (fun x -> Ce_ident x) idl) in
	    let sl_val = List.combine (snd fi.fun_param) el in
	    let sub = mk_subst (sl_ref @ sl_val) in
	    let modif = List.fold_left (fun s x -> IdSet.add x s) (map_idset sub fi.fun_modif) idl in
	    let (sub_ex, sub_ex_post) = 
	      let exfv_pre = ex_fv_can_prop fi.fun_pre in
	      let exfv_post = IdSet.diff (ex_fv_can_prop fi.fun_post) exfv_pre in
	      (mk_subst (List.map (fun id -> (id, Ce_ident(gensym id))) (IdSet.elements exfv_pre)),
	       existentials_rename_sub exfv_post) in
            let post = map_prop (fun e -> sub_ex_post (sub_ex (sub e))) fi.fun_post in
	    let pre = map_prop (fun e -> sub_ex (sub e)) fi.fun_pre in
	    (modif,pre,post) in
	  let go (m,p,q) x =
	    let (m',p',q') = go_fun x in
	    (IdSet.union m' m, can_prop_star p' p, can_prop_star q' q) in
          let (modif,cp1,cp2) = List.fold_left go (IdSet.empty,can_prop_empty,can_prop_empty) par_calls in
	  (* 2. Apply the generic command *)
	  let sub_modif = mk_gensym_garb_subst_list (IdSet.elements modif) in
	  let fr =
	    try prover.find_frame_prop [] spre.se_normal cp1
	    with No_frame_found -> raise (Symb_exe_noframe (spre.se_normal,cp1,"Function call",loc1)) in
	  pp_frame_inferred (loc1,fr);
	  let fr = map_prop sub_modif fr in
	  let cq = can_prop_star fr cp2 in
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_fcall2 (modif,cp1,cp2,s'); can_st_loc=loc1}::c1 ->
	  let sub_modif = mk_gensym_garb_subst_list (IdSet.elements modif) in
	  let fr =
	    try prover.find_frame_prop [] spre.se_normal cp1
	    with No_frame_found -> raise (Symb_exe_noframe (spre.se_normal,cp1,s',loc1)) in
	  pp_frame_inferred (loc1,fr);
	  let fr = map_prop sub_modif fr in
	  let cq = can_prop_star fr cp2 in
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_assume cp; can_st_loc=loc1}::c1 ->
	  let cq = can_prop_star spre.se_normal cp in 
	  symb_execute (rctx,{spre with se_normal=cq},c1,lvs,loc)
      | {can_st_desc=Cst_as_action(c1,rid,cp_ctx,cp0,cq0,modif_body); can_st_loc=loc1}::c2 ->
	  (* Freshen existential variables *)
	  let (cp_ctx,cp0,cq0) =
	    let exfv = IdSet.elements (IdSet.union (ex_fv_can_prop cp0) (IdSet.union (ex_fv_can_prop cq0) (ex_fv_can_prop cp_ctx))) in
	    let sub = mk_subst (List.map (fun id -> (id, Ce_ident(gensym id))) exfv) in
	    (map_prop sub cp_ctx, map_prop sub cp0, map_prop sub cq0) in
	  (* Consider each disjunct separately *)
	  let spost = List.fold_left (fun se_res cf ->
	    (* 1. Calculate the frame [fr] that is not accessed by c1 *)
	    let fr = 
              let cp_shared = try List.assoc rid (snd cf) with Not_found -> assert false in
	      let cp1 = and_cprop_pure cp_shared (fst(fst cf)) in           
	      try prover.find_frame_prop (snd(fst cf)) cp1 cp0
	      with No_frame_found -> raise (Symb_exe_noframe (cp1,cp0,"action entry",loc1)) in
	    (* 2. Check that the frame [fr] contains the context: [fr |- cp_ctx * true] *)
            let _ = 
	      try 
		let sub = existentials_rename_tonormal_sub (ex_fv_can_prop fr) in
		let fr = map_prop sub fr in
		let cp_ctx = map_prop sub cp_ctx in
		prover.find_frame_prop (snd(fst cf)) fr cp_ctx
	      with No_frame_found -> raise (Symb_exe_noframe (fr,cp_ctx,"action entry context",loc1)) in
	    (* 3. Calculate the precondition inside the block *)
	    let cp' = 
	      let scpl' = (rid,fr)::List.remove_assoc rid (snd cf) in
	      can_prop_star {cfl=[(fst cf, scpl')]} cp0 in
	    (* 4. rename vars modified by region body and appear in action post *)
	    let (cp',cq0) =
	      let sub_modif_pairs = 
	        let modif_vars = IdSet.elements (IdSet.inter modif_body (fv_norm_can_prop cq0)) in
	        List.map (fun id -> (id, Ce_ident (gensym id))) modif_vars in
	      let cq0 = map_prop (mk_subst sub_modif_pairs) cq0 in
	      let pl' = List.fold_left (fun res (id,e)->pure_conj_one (mk_EQ(Ce_ident id,e)) res) 
			  pure_true sub_modif_pairs in
	      let cp' = and_cprop_pure cp' pl' in
	      (cp',cq0) in
	    (* 5. Execute the body of the block *)
	    let se = 
	      let fv = fv_can_prop cq0 in
	      let lv = match c2 with [] -> lvs.lv_post | {can_st_lv = lv}::_ -> lv in
	      let lvs' = 
		{ lv_post = IdSet.union lv fv;
	          lv_break = IdSet.union lvs.lv_break fv;
	          lv_cont = IdSet.union lvs.lv_cont fv;
	          lv_return = IdSet.union lvs.lv_return fv } in
	      symb_execute (rctx,{se_false with se_normal = cp'},c1,lvs',loc) in 
	    (* 6. Calculate the postconditions outside the block *)
	    let se = se_map (exit_from_block prover (rid,modif_body,cq0,loc1)) se in
	    se_or se se_res) {spre with se_normal = can_prop_false} spre.se_normal.cfl
	  in
	  symb_execute (rctx,spost,c2,lvs,loc)
      | {can_st_desc=Cst_region(c1,rid,modif,rely); can_st_loc=loc1}::c2 ->
	  let add_inv (uf,scpl) = 
	    if List.exists (fun (x,_)->x=rid) scpl then (uf,scpl) 
            else (uf,(rid,Hashtbl.find renv rid)::scpl) in
	  (* 1. Execute block *)
	  let lvs' = match c2 with [] -> lvs | {can_st_lv = lv}::_ -> {lvs with lv_post = lv} in
	  let spre_inner = {se_false with se_normal={cfl=List.map add_inv spre.se_normal.cfl}} in
	  let spost_inner = symb_execute (rid::rctx,spre_inner,c1,lvs',loc) in 
	  (* 2. Stabilize the postconditions *)
	  let spost_inner =
	    let go = stabilize prover abstraction (rid,modif,rely) in
	    { se_normal = go lvs'.lv_post   spost_inner.se_normal;
	      se_break  = go lvs'.lv_break  spost_inner.se_break;
	      se_cont   = go lvs'.lv_cont   spost_inner.se_cont;
	      se_return = go lvs'.lv_return spost_inner.se_return } in
	  let spost = se_or { spre with se_normal = can_prop_false } spost_inner in
	  symb_execute (rctx,spost,c2,lvs,loc)
      | {can_st_desc=Cst_loop(c1,None); can_st_lv=lv_loop; can_st_loc=loc1}::c2 ->
	  (* Cell to store the exit conditions *)
	  let result = ref se_false in
	  let lvs' = 
	     let lv = match c2 with [] -> lvs.lv_post | {can_st_lv = lv}::_ -> lv in
             { lv_post = lv_loop;  lv_break = lv;
               lv_cont = lv_loop;  lv_return = lvs.lv_return } in
	  let transf cp =
             let se = symb_execute (rctx,{se_false with se_normal=cp},c1,lvs',loc) in
	     result.contents <- se_or result.contents {se with se_normal = can_prop_false; se_cont = can_prop_false};
	     kill_garbage prover lv_loop (can_prop_or se.se_normal se.se_cont) in
	  (* 1. Compute fixpoint *)
	  let _prop_fix = compute_transf_fixpoint abstraction transf spre.se_normal in
	  let spost =
            let (_,cp,_) = abstraction.prop_join can_prop_false can_prop_false result.contents.se_break in
	    { se_normal = cp;
	      se_break  = spre.se_break;
	      se_cont   = spre.se_cont;
	      se_return = can_prop_or result.contents.se_return spre.se_return } in
	  (* 2. Return the result *)
	  symb_execute (rctx,spost,c2,lvs,loc)
      | {can_st_desc=Cst_loop(c1,Some inv); can_st_lv=lv_loop; can_st_loc=loc1}::c2 ->
	  (* 1. Check precondition entails loop invariant *)
	  let fr =
	    try prover.find_frame_prop [] spre.se_normal inv
	    with No_frame_found -> raise (Symb_exe_noframe (spre.se_normal,inv,"loop entry",loc1)) in
	  pp_frame_inferred (loc1,fr);
	  (* 2. Execute the body of the loop *)
	  let lvs' = 
	     let lv = match c2 with [] -> lvs.lv_post | {can_st_lv = lv}::_ -> lv in
             { lv_post = lv_loop;  lv_break = lv;
               lv_cont = lv_loop;  lv_return = lvs.lv_return } in
	  let se =
	    symb_execute (rctx,{se_false with se_normal=inv},c1,lvs',loc) in
	  (* 3. Check postcondition of the loop body entails loop invariant *)
	  let () = 
	    let cq = can_prop_or se.se_cont se.se_normal in
	    let fr' =
	      try prover.find_frame_prop [] cq inv
	      with No_frame_found -> raise (Symb_exe_noframe (cq,inv,"loop exit",loc1)) in
	    if not (is_can_prop_empty fr') then
	      if !Config.allow_leaks then print loc "@[Warning: the loop leaks@ %a@]@." pp_can_prop fr'
	      else raise (Symb_exe_leakmem (fr',loc)) in
	  (* 4. Calculate the postcondition of the loop *)
	  let spost =
	    { se_normal = can_prop_star fr se.se_break;
	      se_break  = spre.se_break;
	      se_cont   = spre.se_cont;
	      se_return = can_prop_or (can_prop_star fr se.se_return) spre.se_return } in
	  symb_execute (rctx,spost,c2,lvs,loc)
     | {can_st_desc=Cst_goto(Cgo_break); can_st_lv=lv}::c2 ->
	  let spost = { spre with se_normal = can_prop_false; 
			se_break = can_prop_or spre.se_normal (kill_garbage prover lv spre.se_break)} in
	  symb_execute (rctx,spost,c2,lvs,loc)
     | {can_st_desc=Cst_goto(Cgo_continue); can_st_lv=lv}::c2 ->
	  let spost = { spre with se_normal = can_prop_false; 
			se_cont = can_prop_or spre.se_normal (kill_garbage prover lv spre.se_cont)} in
	  symb_execute (rctx,spost,c2,lvs,loc)
     | {can_st_desc=Cst_goto(Cgo_return); can_st_lv=lv}::c2 ->
	  let spost = { spre with se_normal = can_prop_false; 
			se_return = can_prop_or spre.se_normal (kill_garbage prover lv spre.se_return)} in
	  symb_execute (rctx,spost,c2,lvs,loc)

in
  let lvs =
    { lv_post = fv2; lv_break = IdSet.empty; lv_cont = IdSet.empty; lv_return = fv2 } in
  let se = symb_execute ([],{se_false with se_normal=cp1},c,lvs,loc) in
  assert (se.se_break = can_prop_false); (* allowed only in a loop *)
  assert (se.se_cont = can_prop_false);  (* allowed only in a loop *)
  can_prop_or se.se_return se.se_normal 


let check_resource prover abstraction ht renv (rid,what) =
  fprintf !Config.formatter "@.Resource %s@." rid;
  try match what with
    | Cri_inv (inv,rely,loc) ->
  	Hashtbl.add renv rid inv;
	let inv' = make_stable prover abstraction (fv_norm_can_prop inv) inv rely in
	let fr =
	  try prover.find_frame_prop [] inv' inv
	  with No_frame_found -> raise (Symb_exe_noframe (inv',inv,"Resource ",loc)) in
	let _ = 
	  if not (is_can_prop_empty fr) && not (!Config.allow_leaks) then 
	    raise (Symb_exe_leakmem (fr,loc)) in
	true
    | Cri_code (fv_post,c,rely,loc) ->
  	Hashtbl.add renv rid can_prop_empty;
	let cq = calculate_post prover abstraction ht renv (can_prop_empty,c,fv_post,loc) in
	let cq = 
	  let fv = IdSet.diff (fv_can_prop cq) fv_post in
	  let sub = mk_gensym_garb_subst_list (IdSet.elements fv) in
	  map_prop sub cq in
	let inv = make_stable prover abstraction fv_post cq rely in
	fprintf !Config.formatter "@.FOUND RESOURCE INVARIANT:@.%a@." pp_can_prop inv;
	Hashtbl.replace renv rid inv; true
  with
    | Symb_exe_noframe (cp1,cp2,s,loc) ->
	print loc "ERROR cannot find frame: %s@.%a@." s pp_can_entailment (cp1,cp2);
	false
    | Symb_exe_leakmem (cp1,loc) ->
	print loc "ERROR memory leak:@.%a@." pp_can_prop cp1;
	false
    | Symb_exe_heap_error -> false

let check_entailment prover abstraction ht renv (fname,(cp1,c,cp2,loc)) =
  fprintf !Config.formatter "@.Function %s@." fname;
  if !Config.verbose1 then
    fprintf !Config.formatter "@.VERIFICATION CONDITION:@.[@[%a@]]@.%a@.[@[%a@]]@.@."
      pp_can_prop cp1 pp_can_cmd c pp_can_prop cp2;
  try 
    (* 1. Execute the body *)
    let cq =
      let fv = fv_can_prop cp2 in
      calculate_post prover abstraction ht renv (cp1,c,fv,loc) in
    if !Config.verbose1 then
      fprintf !Config.formatter "Derived postcondition:@.%a@." pp_can_prop cq;
    (* 2. Check the postcondition *)
    let fr =
      try prover.find_frame_prop [] cq cp2
      with No_frame_found -> raise (Symb_exe_noframe (cq,cp2,"postcondition",loc)) in
    (* 3. Are there any memory leaks? *)
    if not (is_can_prop_empty fr) then
      if !Config.allow_leaks then print loc "Warning: leaking %a@." pp_can_prop fr
      else raise (Symb_exe_leakmem (fr,loc));
    true
  with 
    | Symb_exe_noframe (cp1,cp2,s,loc) ->
	print loc "ERROR cannot find frame: %s@.%a@." s pp_can_entailment (cp1,cp2);
	false
    | Symb_exe_leakmem (cp1,loc) ->
	print loc "ERROR memory leak:@.%a@." pp_can_prop cp1;
	false
    | Symb_exe_heap_error -> false


(** Entry point *)

let check_props prover abstraction ht ril entl =
  let renv = Hashtbl.create 15 in
  let res = List.fold_left (fun res x -> check_resource prover abstraction ht renv x && res) true ril in
  let res = List.fold_left (fun res x -> check_entailment prover abstraction ht renv x && res) res entl in
  res
