open Misc
open Exp

(** Only allow comparisons between integers to catch uses of builtin equality.
  Use Pervasives.(=) on base types only *)
let (=) (x:int) (y:int) = x=y

type link_kind = 
  | SINGLE of string * string * fields (* Node name, component name *)
  | DOUBLE of component * component 
  | XOR of component

type can_spred =
  | Csp_listsegi of link_kind * can_exp * can_exp * can_exp * can_exp * disj_set
  | Csp_node of string * can_exp * fields
  | Csp_tree of component * component * can_exp
  | Csp_indpred of string * can_exp list

(** Spatial part: star-conjuction of can_spred's *)
type can_spatial = can_spred list

type can_uform = can_pure * can_spatial             (** Canonical form without boxes *)
type can_form = can_uform * (string * can_prop) list (** Canonical form *)

and can_prop = {cfl: can_form list}

type can_goto = Cgo_break | Cgo_continue | Cgo_return

type can_stmt =
  { can_st_desc: can_st_desc;
    mutable can_st_lv: IdSet.t;
    can_st_loc: Location.t; }

and can_cmd = can_stmt list

and can_st_desc =
  | Cst_nondet of can_cmd * can_cmd
  | Cst_assign of ident * can_exp
  | Cst_fldlookup of ident * can_exp * string
  | Cst_fldassign of can_exp * string * can_exp
  | Cst_new of ident
  | Cst_dispose of can_exp
  | Cst_pfcall of (ident * string * ident list * can_exp list) list
  | Cst_fcall2 of IdSet.t * can_prop * can_prop * string
  | Cst_assume of can_prop
  | Cst_region of can_cmd * string * Misc.IdSet.t * (string * can_prop * can_prop * can_prop) list
  | Cst_as_action of can_cmd * string * can_prop * can_prop * can_prop * Misc.IdSet.t
  | Cst_loop of can_cmd * can_prop option
  | Cst_goto of can_goto

type can_entailment = can_prop * can_cmd * can_prop * Location.t

type fun_info =
    { fun_param: ident list * ident list;
      fun_modif: IdSet.t;
      fun_pre: can_prop;
      fun_post: can_prop;
      fun_loc: Location.t }

exception No_frame_found
exception Junk

type prover =
   { find_ptsto : string -> can_exp -> can_spatial -> fields * can_spatial;
     expose_ptsto : can_prop -> can_exp -> can_prop option;
     normal_form_prop : can_spatial -> equalities -> can_prop -> equalities * can_prop;
     normal_can_form : can_form -> can_form option;
     entails_form : can_form -> can_form -> bool;
     find_frame_prop : can_spatial -> can_prop -> can_prop -> can_prop;
   }

type abstraction =
  {
     uform_abs : can_uform list -> can_uform list;
     uform_join : can_uform list -> can_uform list -> can_uform list -> (can_uform list * can_uform list * can_uform list);

     (** [prop_abs] returns a [cp'] belonging to a smaller domain such that [cp |- cp'] *)
     prop_abs : can_prop -> can_prop;
     prop_join : can_prop -> can_prop -> can_prop -> (can_prop * can_prop * can_prop);
     
  }

type res_init = 
  | Cri_inv of can_prop * (string * can_prop * can_prop * can_prop) list * Location.t
  | Cri_code of IdSet.t * can_cmd * (string * can_prop * can_prop * can_prop) list * Location.t

(* Option monad *)
let (>>=) x f =
  match x with
  | None -> None
  | Some v -> f v

let compare_lkind ?(cmp_id=compare_ident) k1 k2 = match k1,k2 with
  | SINGLE(s1,s2,cel), SINGLE(s1',s2',cel') ->
      let n = compare s1 s1' in if n<>0 then n else
      let n = compare s2 s2' in if n<>0 then n else
      compare_fields cel cel'
  | SINGLE _, _ -> -1
  | _, SINGLE _ -> 1
  | DOUBLE(s1,s2), DOUBLE(s1',s2') ->
      let n = compare s1 s1' in if n<>0 then n else
      compare s2 s2'
  | DOUBLE _, _ -> -1
  | _, DOUBLE _ -> 1
  | XOR(s1), XOR(s1') -> compare s1 s1'

let compare_can_spred ?(cmp_id=compare_ident) sp1 sp2 = match sp1,sp2 with
  | Csp_listsegi (k,e1,e2,e3,e4,ds), Csp_listsegi(k',e1',e2',e3',e4',ds') ->
      let n = compare_lkind k k' in if n<>0 then n else
      let n = compare_exp e1 e1' in if n<>0 then n else
      let n = compare_exp e2 e2' in if n<>0 then n else
      let n = compare_exp e3 e3' in if n<>0 then n else
      let n = compare_exp e4 e4' in if n<>0 then n else
      compare_dset ds ds'
  | Csp_listsegi _, _ -> -1
  | _, Csp_listsegi _ -> 1
  | Csp_node (s1,e1,cel1), Csp_node (s2,e2,cel2) ->
      let n = Pervasives.compare s1 s2 in if n<>0 then n else
      let n = compare_exp e1 e2 in if n<>0 then n else 
      compare_fields cel1 cel2
  | Csp_node _, _ -> -1
  | _, Csp_node _ -> 1
  | Csp_tree (t1,t2,e), Csp_tree(t1',t2',e') ->
      let n = compare t1 t1' in if n<>0 then n else
      let n = compare t2 t2' in if n<>0 then n else
      compare_exp e e'
  | Csp_tree _, _ -> -1
  | _, Csp_tree _ -> 1
  | Csp_indpred (s1, el1), Csp_indpred (s2,el2) ->
      let n = compare s1 s2 in if n<>0 then n else
      compare_exp_list el1 el2

let rec compare_can_spatial ?(cmp_id=compare_ident) sl1 sl2 = match sl1,sl2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | sp1::sl1', sp2::sl2' ->
      let n = compare_can_spred sp1 sp2
      in if n<>0 then n
	else compare_can_spatial sl1' sl2'

and compare_uform ?(cmp_id=compare_ident) (pl1,sl1) (pl2,sl2) =
  let n = compare_pure pl1 pl2
  in if n<>0 then n
    else compare_can_spatial sl1 sl2

let rec compare_string_prop ?(cmp_id=compare_ident) ((s1:string),p1) (s2,p2) =
  let n = compare s1 s2
  in if n<>0 then n
    else compare_prop p1 p2

and compare_string_prop_list ?(cmp_id=compare_ident) spl1 spl2 = match spl1,spl2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | x1::spl1', x2::spl2' ->
      let n = compare_string_prop x1 x2
      in if n<>0 then n
	else compare_string_prop_list spl1' spl2'

and compare_can_form ?(cmp_id=compare_ident) ((cu1,spl1): can_form) ((cu2,spl2): can_form) =
  let n = compare_uform cu1 cu2
  in if n<>0 then n
    else compare_string_prop_list spl1 spl2

and compare_prop ?(cmp_id=compare_ident) cp1 cp2 =
  let rec f cfl1 cfl2 = match cfl1,cfl2 with
    | [],[] -> 0
    | [],_ -> -1
    | _,[] -> 1
    | cf1::cfl1', cf2::cfl2' ->
	let n = compare_can_form cf1 cf2
	in if n<>0 then n
	  else f cfl1' cfl2'
  in f cp1.cfl cp2.cfl

let rec compare_component_exp ?(cmp_id=compare_ident) ((s1:component),e1) (s2,e2) =
  let n = compare s1 s2
  in if n<>0 then n
    else compare_exp e1 e2

let compare_ident_exp ?(cmp_id=compare_ident) (i1,e1) (i2,e2) =
  let n = compare_ident i1 i2
  in if n<>0 then n
    else compare_exp e1 e2

let mem_ident i l =
  List.exists (fun x -> equal_ident x i) l

let mem_exp y l =
  List.exists (fun x -> equal_exp x y) l

let mem_string_prop y l =
  List.exists (fun x -> compare_string_prop x y = 0) l

let spat_empty = []
let spat_one sp = [sp]
let spat_star_one sp sl = sp::sl
let spat_star sl1 sl2 = sl1@sl2
let spat_fold go sl e = List.fold_left (fun x y -> go y x) e sl
let spat_remove sl x = List.filter (fun y -> compare_can_spred y x <> 0) sl

(**** pretty printing ****)
open Format

let pp_string f s = fprintf f "%s" s

let rec pp_listsep pp s1 s2 f = function
  | [] -> fprintf f "%s" s2
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a%s %a" pp x s1 (pp_listsep pp s1 s2) l

let rec pp_seq pp f = function
  | [] -> ()
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a,%a" pp x (pp_seq pp) l

let pp_disjoint_set f ds = match dset_to_list ds with
  | [] -> fprintf f ""
  | el -> fprintf f ",{%a}" (pp_seq pp_can_exp) el

let pp_comp_exp_list =
  let pp f (c,e) = fprintf f "%s:%a" c pp_can_exp e in 
  pp_seq pp

let pp_can_spred f = function
  | Csp_listsegi (SINGLE(snode,sfun,cel),e1,f1,e2,_,ds) ->
      let (what,ds) = if dset_mem f1 ds then ("lseg",dset_remove f1 ds) else ("lsegi",ds) in
      let el = if Pervasives.(=) sfun "@zero" then [e1;f1] else [e1;f1;e2] in
      let s = (if Pervasives.(=) snode ("."^Config.list_link_tag) then "" else snode^";") in
      (match fld_to_list cel with
	| [] -> fprintf f "%s(%s%a%a)" what s (pp_seq pp_can_exp) el pp_disjoint_set ds
	| cel -> fprintf f "%s[%a](%s%a%a)" what pp_comp_exp_list cel s (pp_seq pp_can_exp) el pp_disjoint_set ds)
  | Csp_listsegi (DOUBLE(t1,t2),e1,f1,e2,f2,el) ->
      fprintf f "dlsegi(%s,%s;%a,%a,%a,%a%a)"
      t1 t2 pp_can_exp e1 pp_can_exp f1 pp_can_exp e2 pp_can_exp f2 pp_disjoint_set el
  | Csp_listsegi (XOR t1,e1,f1,e2,f2,el) ->
      fprintf f "xlsegi(%s%a,%a,%a,%a%a)"
      (if Pervasives.(=) t1 Config.dl_Llink_tag then "" else t1^";") pp_can_exp e1 pp_can_exp f1 pp_can_exp e2 pp_can_exp f2 pp_disjoint_set el
  | Csp_node (s, e1, cel) ->
      begin match fld_to_list cel with
	| [] -> fprintf f "@[%a|->_@]" pp_can_exp e1
	| cel -> fprintf f "@[%a|->%a@]" pp_can_exp e1 pp_comp_exp_list cel
      end 
  | Csp_tree (t1,t2,e) ->
      fprintf f "tree(%s%a)"
	(if Pervasives.(=) (t1,t2) Config.tree_link_tags then "" else t1^","^t2^";") pp_can_exp e
  | Csp_indpred (id,el) ->
      fprintf f "%s(%a)" id (pp_seq pp_can_exp) el


let rec pp_starsep pp s2 f = function
  | [] -> fprintf f "%s" s2
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a@ * %a" pp x (pp_starsep pp s2) l

let rec pp_orsep pp f = function
  | [] -> fprintf f "false"
  | [x] -> fprintf f "@[%a@]" pp x
  | x::l -> fprintf f "@[%a@]@ || %a" pp x (pp_orsep pp) l

let pp_pure_list = pp_starsep pp_can_atom ""
let pp_spatial_list = pp_starsep pp_can_spred "emp"

let get_alloc (pl,sl) = map_partial (function Csp_node(_,e,_)-> Some e | _ -> None) sl

let pp_simp alloc pl =
(*  let alloc = Ce_num 0 :: alloc in
  List.filter (function NEQ(e1,e2) -> not (mem_exp e1 alloc && mem_exp e2 alloc) | _ -> true)*) pl

let pp_can_uform_a alloc f (pl,sl) = 
  let pl = pure_to_list pl in
  let pl = if !Config.pretty_print then pp_simp alloc pl else pl in
  match (pl,sl) with
    | ([],[]) -> fprintf f "emp"
    | (_::_,[]) -> pp_pure_list f pl
    | ([],_::_) -> pp_spatial_list f sl
    | _ -> fprintf f "%a@ * %a" pp_pure_list pl pp_spatial_list sl

let rec pp_box_form_a alloc f (s,cp) = fprintf f "%s:[@[%a@]]" s (pp_can_prop_a alloc) cp

and pp_can_form_a alloc f (uf,scpl) =
  let alloc = get_alloc uf @ alloc in
  match (uf,scpl) with
    | (_,[]) -> pp_can_uform_a alloc f uf
    | _ -> fprintf f "@[<hv>@[%a@]@ * %a@]" (pp_can_uform_a alloc) uf (pp_starsep (pp_box_form_a alloc) "") scpl

and pp_can_prop_a alloc f cp =
  fprintf f "@[<hv>%a@]" (pp_orsep (pp_can_form_a alloc)) cp.cfl

let pp_can_form f cf = pp_can_form_a [] f cf
let pp_can_prop f cp = pp_can_prop_a [] f cp

let rec pp_can_stmt f st = 
  let go f st = match st.can_st_desc with
  | Cst_nondet (c1,c2) ->
      fprintf f "@[nondet@;  %a@;or@;  %a@]"
      pp_can_cmd c1 pp_can_cmd c2
  | Cst_assign (x,e) ->
      fprintf f "%a=%a;" pp_ident x pp_can_exp e
  | Cst_fldlookup (x,e,t) ->
      fprintf f "%a=%a->%s;" pp_ident x pp_can_exp e t
  | Cst_fldassign (e1,t,e2) ->
      fprintf f "%a->%s=%a;"
      pp_can_exp e1 t pp_can_exp e2
  | Cst_new (x) ->
      fprintf f "new(%a);" pp_ident x
  | Cst_dispose (e) ->
      fprintf f "dispose(%a);" pp_can_exp e
  | Cst_pfcall [(v,s,vl,el)] ->
      fprintf f "%a = %s(@[<hov 2>%a;@ %a@]);" pp_ident v s 
	(pp_seq pp_ident) vl (pp_seq pp_can_exp) el
  | Cst_pfcall xl ->
      fprintf f "[@par@;  %a@]" pp_can_cmd 
	(List.map (fun x -> {can_st_desc = Cst_pfcall [x]; 
			     can_st_lv = st.can_st_lv;
			     can_st_loc = st.can_st_loc}) xl)
  | Cst_fcall2 (modif,cp1,cp2,s) ->
      fprintf f "fcall(@[<hv>{%a},@[<hov 2>%a@],@ @[<hov 2>%a@]@]);"
      pp_idset modif pp_can_prop cp1 pp_can_prop cp2
  | Cst_assume (cp) ->
      fprintf f "assume(%a);" pp_can_prop cp
  | Cst_region (c,rid,modif,rely) ->
      fprintf f "@[<hv 2>with %s@ %a@ %a@];" rid pp_can_cmd c
	pp_idset modif
  | Cst_as_action(c,rid,cr,cp,cq,modif) ->
      fprintf f "@[<hv 2>do@ %a@ as %a@ | %a@ --> %a@];" pp_can_cmd c
	pp_can_prop cr pp_can_prop cp pp_can_prop cq
  | Cst_loop (c,None) ->
      fprintf f "@[<hv 2>loop@ %a;@]" pp_can_cmd c 
  | Cst_loop (c,Some cp) ->
      fprintf f "@[<v 2>loop@ @[invariant:@ %a@]@ %a@]" pp_can_prop cp pp_can_cmd c 
  | Cst_goto (Cgo_break)    -> fprintf f "break;"
  | Cst_goto (Cgo_continue) -> fprintf f "continue;"
  | Cst_goto (Cgo_return)   -> fprintf f "return;"
  in
    (*fprintf f "[LV: %a]@;%a" pp_idset st.can_st_lv go st*)
    go f st

and pp_can_cmd f sl =
  let rec go f = function 
  | [] -> fprintf f "{}"
  | [x] -> fprintf f "%a" pp_can_stmt x
  | x::l -> fprintf f "%a@;%a" pp_can_stmt x go l
  in
    fprintf f "@[<v>%a@]" go sl

let pp_can_entailment f (cp1,cp2) =
  fprintf f "@[<hv>@[<hov 2>%a@]@ |-@ @[<hov 2>%a@]@]"
    pp_can_prop cp1 pp_can_prop cp2

let pp_sub f sub =
  (pp_seq (fun f (x,e) -> fprintf f "%s=%a" (string_of_ident x) pp_can_exp e)) f sub

let pp_inst f inst =
  if !Config.verbose1 then
    if inst <> []
    then fprintf f "Instantiations inferred: %a@." pp_sub inst

(** default values *)
let can_atom_false = mk_NEQ (Ce_num 0, Ce_num 0)
let can_atom_true = mk_NEQ (Ce_num 0, Ce_num 1)
let can_uform_false = (pure_one can_atom_false,[])
let can_form_false = (can_uform_false,[])
let can_prop_pure pl = {cfl=[((pl,[]),[])]}
let can_prop_empty = can_prop_pure (pure_true)
let can_prop_false = {cfl=[]}

let remove_uform_duplicates l =
  let l2 = List.stable_sort compare_uform l in
  let rec f = function
    | x::y::xs -> if compare_uform x y = 0 then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2

let remove_can_form_duplicates l =
  let l2 = List.stable_sort compare_can_form l in
  let rec f = function
    | x::y::xs -> if compare_can_form x y = 0 then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2

let normalize_can_uform (pl,sl) =
  match normalize_pure pl with
    | None -> None
    | Some pl' -> Some (pl',sl)

let normalize_can_form (uf,scpl) =
  match normalize_can_uform uf with
    | None -> None
    | Some uf' -> Some (uf', scpl)

let normalize_can_prop cp =
  {cfl = map_partial normalize_can_form cp.cfl}



(* substitutions *)
type subst = can_exp -> can_exp 

let rec mk_subst (l: (ident * can_exp) list) e =
  let rec f_e e = match e with
    | Ce_num n -> e
    | Ce_ident i ->
	let rec f = function
	  | [] -> e
	  | (i',e')::il -> if equal_ident i i' then e'
	    else f il
	in f l
    | Ce_fun (i,el) -> mk_fun (i, List.map f_e el)
    | Ce_grp(op,n,nel) -> mk_grp(op,n,List.map (fun(n,e)->(n, f_e e)) nel)
  in f_e e

let mk_single_subst (i,e) =
  mk_subst [(i,e)]

let mk_gensym_garb_subst i =
  mk_single_subst (i, Ce_ident (gensym_garb i))

let gensym_subst_list is_garb l =
  let l' = List.map (fun i -> (i,Ce_ident ((if is_garb then gensym_garb else gensym) i))) l in
  mk_subst l'

let mk_gensym_subst_list =
  gensym_subst_list false

let mk_gensym_garb_subst_list =
  gensym_subst_list true

let map_idset f_e ss =
  let f i = match f_e (Ce_ident i) with
    | Ce_ident x -> x
    | _ -> assert false
  in IdSet.fold (fun i s -> IdSet.add (f i) s) ss IdSet.empty

let rec map_spred f_e = function
  | Csp_listsegi (k,e1,f1,e2,f2,el) -> Csp_listsegi (k, f_e e1, f_e f1, f_e e2, f_e f2, map_dset f_e el)
  | Csp_node (s,e,cel) -> Csp_node (s,f_e e, map_fld f_e cel)
  | Csp_tree (t1,t2,e) -> Csp_tree (t1,t2, f_e e)
  | Csp_indpred (id,el) -> Csp_indpred (id, List.map f_e el)

let map_uform f_e (pl,sl) =
  match normalize_pure (map_pure f_e pl) with
  | None -> None
  | Some pl' -> Some (pl', List.map (map_spred f_e) sl)

(*let rec map_spred f_e = function
  | Csp_listsegi (k,e1,f1,e2,f2,el) -> Some (Csp_listsegi (k, f_e e1, f_e f1, f_e e2, f_e f2, map_dset f_e el))
  | Csp_node (s,e,cel) ->
      begin match f_e e with
	| Ce_num 0 -> None 
	| e' -> Some (Csp_node (s,e', map_fld f_e cel))
      end
  | Csp_tree (t1,t2,e) -> Some (Csp_tree (t1,t2, f_e e))
  | Csp_indpred (id,el) -> Some (Csp_indpred (id, List.map f_e el))

let map_uform f_e (pl,sl) =
  match normalize_pure (map_pure f_e pl) with
  | None -> None
  | Some pl' -> Misc.map_option (map_spred f_e) sl >>= fun sl' -> Some(pl', sl')*)

let rec map_form f_e (uf, scpl) =
  map_uform f_e uf >>= fun uf' ->
  let scpl' = List.map (fun (s,cp) -> (s,map_prop f_e cp)) scpl in
  Some (uf',scpl')
  (*(sub_can_uform sub uf, List.map (fun (s,cp) -> (s,sub_can_prop sub cp)) scpl)*)

and map_prop f_e cp =
  {cfl = map_partial (map_form f_e) cp.cfl}

let map_ip_body f_e body =
  let f (pl,cf) = 
    normalize_pure (map_pure f_e pl) >>= fun pl' ->
    map_form f_e cf >>= fun cf' ->
    Some (pl',cf') in
  map_partial f body


let (++) = IdSet.union

let fv_can_spred = function
  | Csp_listsegi (_,ce1,ce2,ce3,ce4,ds) ->
      (fv_can_exp ce1) ++ (fv_can_exp ce2) ++ (fv_can_exp ce3) ++ (fv_can_exp ce4) ++ (fv_dset ds)
  | Csp_node (_,ce, fld) -> fv_can_exp ce ++ fv_fld fld
  | Csp_tree (_,_,ce) -> fv_can_exp ce
  | Csp_indpred (i,el) -> List.fold_left (fun s ce -> fv_can_exp ce ++ s) IdSet.empty el

let fv_can_spred_fromTID = function
  | Csp_node (_,ce, fld) ->
      let go_fld (res,fv) (c,e) = 
	 let idset = fv_can_exp e in
	 if IdSet.mem tid_var idset then
	   if Config.is_lock_field c then (true,fv)
	   else (true,idset++fv)
	 else if Config.is_lock_field c then (res,fv)
	 else (res, idset++fv) in
      let (res,fv) = List.fold_left go_fld (false,fv_can_exp ce) (fld_to_list fld) in
      if res then fv else IdSet.empty
  | Csp_listsegi _
  | Csp_tree _
  | Csp_indpred _ ->
      IdSet.empty

let fv_can_uform (pl,sl) =
  pure_fold (fun ca x -> fv_can_atom ca ++ x) pl
    (List.fold_left (fun x sp -> fv_can_spred sp ++ x) IdSet.empty sl)

let fv_can_uform_fromTID (pl,sl) =
  let fvl = List.map fv_can_spred_fromTID sl
  in List.fold_left (++) IdSet.empty fvl

let rec fv_can_form (uf,scpl) =
  let fv1 = fv_can_uform uf
  and fvl2 = List.map (fun (_,x)->fv_can_prop x) scpl
  in List.fold_left (++) fv1 fvl2

and fv_can_prop cp =
  List.fold_left (++) IdSet.empty (List.map fv_can_form cp.cfl)

let rec fv_can_form_fromTID (uf,scpl) =
  let fv1 = fv_can_uform_fromTID uf
  and fvl2 = List.map (fun (_,x)->fv_can_prop x) scpl
  in List.fold_left (++) fv1 fvl2

and fv_can_prop_fromTID cp =
  List.fold_left (++) IdSet.empty (List.map fv_can_form_fromTID cp.cfl)


(* Existential free variables only *)
let abs_fv_can_uform (pl,sl) = 
  let rec go_exp fv e = match e with
    | Ce_num _ -> fv
    | Ce_ident i -> if is_existential i && not (mem_ident i fv) then i::fv else fv
    | Ce_fun (i,el) -> List.fold_left go_exp fv el
    | Ce_grp(op,n,nel) -> List.fold_left (fun fv (_,e) -> go_exp fv e) fv nel in
  let go_spred fv sp = match sp with
    | Csp_listsegi (_,ce1,ce2,ce3,ce4,el) -> List.fold_left go_exp fv (ce1::ce2::ce3::ce4::(dset_to_list el))
    | Csp_node (_, ce, ccel) -> List.fold_left (fun fv (_,e) -> go_exp fv e) (go_exp fv ce) (fld_to_list ccel)
    | Csp_tree (_,_,ce) -> go_exp fv ce
    | Csp_indpred (i,el) -> List.fold_left go_exp fv el in
  let fv = List.fold_left go_spred [] sl in
  let fv = pure_fold (fold_atom (fun e v -> go_exp v e)) pl fv in
  List.rev fv


(** unify [e1] and [e2] instantiating existential variables in [e2]. *)
let unify_exp inst e1 e2 = 
  let rec go (inst,pl) e1 e2 = match (e1,e2) with
    | e1, Ce_ident i2 ->
      (try 
	let (_,e3) = List.find (fun (x,_)-> equal_ident x i2) inst in
	go (inst,pl) e1 e3
      with Not_found -> 
	if is_existential i2 then
	  let sub = mk_single_subst (i2,e1) in
	  let inst = List.map (fun (t,e) -> (t,sub e)) inst in
	  Some ((i2,e1)::inst, pl)
	else if equal_exp e1 e2 then Some (inst,pl)
	else None)
    | Ce_num n1, Ce_num n2 -> if n1 = n2 then Some (inst,pl) else None
(*    | Ce_fun (s1,el1), Ce_fun (s2,el2) ->
	if Pervasives.(=) s1 s2 then go_list (inst,pl) el1 el2 else None *)
    | e1, e2 -> if equal_exp e1 e2 then Some (inst,pl) else None
  and go_list res el1 el2 = match (el1,el2) with
    | [], [] -> Some res
    | [],_::_ -> None 
    | _::_,[] -> None
    | e1::el1,e2::el2 -> 
	go res e1 e2 >>= fun res ->
	go_list res el1 el2
  in
  go (inst,pure_true) e1 e2 

let can_spred_leq inst csp1 csp2 = match (csp1,csp2) with
  | (Csp_listsegi(SINGLE(s1,s2,cel),e1,f1,e2,f2,el),
     Csp_listsegi(SINGLE(s1',s2',cel'),e1',f1',e2',f2',el')) ->
      if Pervasives.(=) s1 s1' && Pervasives.(=) s2 s2' && equal_exp e1 e1' && dset_subset el' el then
	if equal_exp f1 f1' then unify_exp inst e2 e2'
        else if equal_exp e2 e2' then unify_exp inst f1 f1'
	else None
      else None
  | (Csp_listsegi(k1,e1,f1,e2,f2,el), Csp_listsegi(k1',e1',f1',e2',f2',el')) ->
      if Pervasives.(=) k1 k1' && equal_exp e1 e1' && equal_exp f1 f1' 
	&& equal_exp e2 e2' && equal_exp f2 f2' && dset_subset el' el 
	then Some (inst,pure_true)
	else None
  | (csp1,csp2) -> 
	if compare_can_spred csp1 csp2 = 0 then Some (inst,pure_true) else None


(* TODO *)
let can_form_star ((pl1,sl1),scpl1) ((pl2,sl2),scpl2) = 
  let scpl',scpl1' = List.partition (fun(x,_)-> List.for_all (fun (y,_)-> x <> y) scpl2) scpl1 in
  assert (List.for_all (fun x-> mem_string_prop x scpl2) scpl1');
  aggr_norm_pure (pure_conj pl1 pl2) >>= fun pl'->
  Some ((pl', sl1@sl2), scpl'@scpl2)

(* Keep cp_star's in a canonical form: 
   - right-associative Cp_star,
   - combine Cp_base's and bring them leftmost. 
Assuming cp1 and cp2 are in canonical form, put cp1*cp2 in canonical form. *)
let can_prop_star cp1 cp2 =
  let f res cf1 = (map_partial (can_form_star cf1) cp2.cfl)@res
  in {cfl = List.fold_left f [] cp1.cfl}

let and_cprop_pure cp pl = can_prop_star cp (can_prop_pure pl)

let remove_can_form_duplicates l =
  let l2 = List.stable_sort compare_can_form l in
  let rec f = function
    | x::y::xs -> if compare_can_form x y = 0 then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2

let can_prop_or cp1 cp2 =
  {cfl = remove_can_form_duplicates (cp1.cfl@cp2.cfl)}

let rec is_pure cp =
  List.for_all (function ((_,[]),_)->true | _ -> false) cp.cfl

let is_can_prop_empty = is_pure

let can_prop_spred sl = {cfl = [((pure_true,sl),[])]}

(*start of wand elim code *)
exception Undef_wand_elim of string

(* Assuming pl, distribute  _ /\ not(x |-> * true) over a list of spatial assertions *)
let notloc x pl sl =
  List.fold_left 
    (fun ufo s ->
      match ufo with
	| None -> None
	| Some (pl',sl') ->
	    match s with 
	      | Csp_node(nid,y,fields) ->
		  (match mk_NEQ_exn (y,x) with
		    | None -> None
		    | Some pl1 -> Some ((pure_conj pl1 pl',Csp_node(nid,y,fields)::sl')))
	      | Csp_listsegi(k,e1,e2,e3,e4,ds) -> 
		  if dset_mem e1 ds && mem_atom (mk_NEQ(e1,e2)) pl 
		  then None
		  else Some ((pl',Csp_listsegi(k,e1,e2,e3,e4,dset_add x ds)::sl'))
	      | _ -> raise (Undef_wand_elim "Cannot use tree or inductive predicates on right of -*E"))
    (Some (pure_true,[]))
    sl

(* returns x |-> nid(fields) -*E s, where s is an spred. *)
let rec wande_ptsto_spred nid x fields s =
  let mk_listseg sfun (k,e1,f1,e2,f2,ds) = 
    if equal_exp e1 f1 && dset_mem f1 ds then
      match mk_EQ_exn(e2,mk_fun(sfun,[])) with 
	| None -> {cfl=[]}
	| Some pl -> can_prop_pure pl
    else {cfl = [((pure_true,[Csp_listsegi(k,e1,f1,e2,f2,ds)]),[])]} in
  match s with 
      | Csp_listsegi (SINGLE(snode,sfun,flds),e1,e2,e3,_,el) ->
	  let t = 
	    if snode <> "" && Pervasives.(=) (String.get snode 0) '.' then
	      String.sub snode 1 (String.length snode - 1)
	    else assert false (* TODO *) in
	  let e = match getcomponent t fields with None -> Ce_ident (gensym (create_ident "_TAIL")) | Some e -> e in
	  begin match (mk_NEQ_exn(Ce_num 0,x),map_option (fun y-> mk_NEQ_exn(x,y)) (dset_to_list el),components_equal fields flds) with
	    | (Some pl0, Some pll, Some pl') ->
	        let e3a = mk_fun(sfun, [Ce_ident (gensym (create_ident "_VAL"))]) in
	        let e3b = mk_fun(sfun, [Ce_ident (gensym (create_ident "_VAL"))]) in
	        let e3c = try fld_get "val" fields with Not_found -> Ce_ident (gensym (create_ident "_VAL")) in
	        let e3d = 
	  	if equal_exp x e1 then mk_fun(sfun,[mk_fun("@item",[e3c]);e3b])
	  	else mk_fun(sfun,[e3a;mk_fun("@item",[e3c]);e3b]) in
	        (match mk_EQ_exn(e3, e3d) with
	  	| None -> {cfl=[]} | Some pl1 ->
	        and_cprop_pure 
	  	  (can_prop_star
		    (mk_listseg sfun (SINGLE(snode,sfun,flds),e1,x,e3a,Ce_num 0,dset_add x el))
	  	    (mk_listseg sfun (SINGLE(snode,sfun,flds),e,e2,e3b,Ce_num 0,dset_add x el)))
		  (pure_conj pl1 (pure_conj pl0 (List.fold_left (fun x e -> pure_conj e x) pl' pll))))
	    | _ -> {cfl=[]}
	  end
      | Csp_listsegi (_,_,_,_,_,_) -> raise (Undef_wand_elim "Cannot use dllseg or xlseg on right of -*E")
      | Csp_node (nid',e,fields') ->
	  if nid'<>nid then {cfl=[]} else
	  (match (mk_NEQ_exn(Ce_num 0,x), mk_EQ_exn(e,x), components_equal fields fields') with
	    | (Some pl1, Some pl2, Some pl3) ->
		can_prop_pure (pure_conj pl1 (pure_conj pl2 pl3))
	    | _ -> {cfl=[]})
      | Csp_tree _ -> raise (Undef_wand_elim "Cannot use tree on right of -*E")
      | Csp_indpred _ -> raise (Undef_wand_elim "Cannot use inductive predicate on right of -*E")

(* returns x|->nid(fields) -*E cf, where sl is a list of spreads *)
let wande_ptsto_cform (nid,x,fields) ((pl,sl),scpl) =
  let my_star ufo f = match ufo with
    | None -> {cfl=[]}
    | Some (pl,sl) ->
	let cfl = (f()).cfl in
	{cfl = (List.map (fun ((pl',sl'),scpl') -> ((pure_conj pl pl',sl@sl'),scpl')) cfl)} in
  let rec go = function
    | [] -> {cfl=[]}
    | s::sl ->
       can_prop_or 
        (my_star (notloc x pl sl) (fun()->wande_ptsto_spred nid x fields s))
	(my_star (notloc x pl [s]) (fun()->go sl))
  in
    if !Config.allow_leaks then 
      (* might also remove x|->fields from the implicit ``* true'' *)
      can_prop_or 
	(match notloc x pl sl with 
	  | None -> {cfl=[]} 
	  | Some (pl',sl') ->{cfl=[((pure_conj pl' pl,sl'),scpl)]}) 
	(can_prop_star (go sl) {cfl = [((pl,[]),scpl)]})
    else
      can_prop_star (go sl) {cfl = [((pl,[]),scpl)]}

(* returns  x|->nid(fields) -*E cp *)
let rec wande_ptsto_cp (nid,x,fields) cp = 
  let cpl = List.map (wande_ptsto_cform (nid,x,fields)) cp.cfl in
  let cfl = List.concat (List.map (function cp -> cp.cfl) cpl) in
  {cfl = remove_can_form_duplicates cfl}

(* return  lseg(SINGLE(node,@zero,[]),e1,f1,0,ds) -*E cf *)
let rec wande_lseg_cform (node,e1,f1,ds) ((pl,sl),scpl) =
  match eq_add_list empty_eq pl with 
   | None -> {cfl=[]}
   | Some eq -> 
      let (e1,f1) = let sub = eq_to_sub eq in (sub e1, sub f1) in
  if equal_exp e1 (Ce_num 0) then 
    match mk_EQ_exn(f1,Ce_num 0) with 
      | None -> {cfl=[]}
      | Some pl_new -> {cfl=[((pure_conj pl_new pl,sl),scpl)]}
  else if equal_exp e1 f1 && dset_mem e1 ds then 
    {cfl=[((pl,sl),scpl)]}
  else (* The rest *)
  let t = 
    if node <> "" && Pervasives.(=) (String.get node 0) '.' then
      String.sub node 1 (String.length node - 1)
    else assert false (* TODO *) in
  let cannot_match sp = match sp with
    | Csp_node (_,e,cel) -> is_uneq pl (e1,e)
    | Csp_listsegi (SINGLE _,_,_,_,_,ds) -> dset_mem e1 ds
    | _ -> false
  in
  let rec go sl0 sl1 = match sl1 with
    | [] ->
	Format.printf "WARNING: Cannot compute lseg(%a,%a) -*E@.%a@." pp_can_exp e1 pp_can_exp f1 pp_can_form ((pl,sl),scpl);
	raise (Undef_wand_elim "Did not manage to eliminate lseg on right of -*E")
    | Csp_node(nid,e,fields)::sl1 when equal_exp e e1 ->
	let f2 = try fld_get t fields with Not_found -> Ce_ident (gensym (create_ident "_TAIL")) in
	begin match mk_EQ_exn(f1,f2) with 
	  | None -> wande_lseg_cform (node,f2,f1,ds) ((pl,List.rev_append sl0 sl1),scpl)
	  | Some pl_new ->
	      if pure_is_true pl_new then
		 {cfl=[((pl,List.rev_append sl0 sl1),scpl)]}
	      else
		can_prop_or {cfl=[((pure_conj pl_new pl,List.rev_append sl0 sl1),scpl)]}
          	  (wande_lseg_cform (node,f2,f1,ds) ((pl,List.rev_append sl0 sl1),scpl))
	end
    | Csp_listsegi(SINGLE(snode,_,fld),e,f2,_,_,ds2)::sl1 when Pervasives.(=) snode node && equal_exp e e1 ->
        begin match mk_EQ_exn(f1,f2) with 
	 | Some pl' when pure_is_true pl' -> {cfl=[((pl,List.rev_append sl0 sl1),scpl)]}
	 | _ -> can_prop_or 
	  {cfl=[((pl, Csp_listsegi(SINGLE(snode,"@zero",fld),f1,f2,Ce_num 0,Ce_num 0,ds2) :: List.rev_append sl0 sl1),scpl)]}
	  (wande_lseg_cform (node,f2,f1,ds) ((pl,List.rev_append sl0 sl1),scpl))
	end
    | s::sl1 -> go (s::sl0) sl1 in
  if List.for_all cannot_match sl then
    match mk_EQ_exn(e1,f1) with
      | None -> {cfl=[]}
      | Some pl_new -> {cfl=[((pure_conj pl_new pl,sl),scpl)]}
  else go [] sl


let rec wande_lseg_cp data cp = 
  let cpl = List.map (wande_lseg_cform data) cp.cfl in
  let cfl = List.concat (List.map (function cp -> cp.cfl) cpl) in
  {cfl = remove_can_form_duplicates cfl}

(* return  sl1 -*E cf2 *)
let rec wande_base_cform sl1 (uf2,scpl2) = 
  (* returns e|->nid(fields) -*E sl2  such that sl2 contains e|->_ *)
  let rec go0 nid e fields sl0 = function
    | (_, []) -> assert false (* Should never happen *)
    | (pl2, Csp_node(nid', x, fields')::sl2) when Pervasives.(=) nid nid' && equal_exp x e -> 
	components_equal fields fields' >>= fun pl ->
	Some (pure_conj pl pl2, List.rev_append sl0 sl2)
    | (pl2, x::sl2) -> go0 nid e fields (x::sl0) (pl2, sl2) in
  let rec go_frame sl1 uf2 = match sl1 with
    | [] -> Some uf2
    | Csp_node(nid,e,cel)::sl1 ->
	go0 nid e cel [] uf2 >>= fun uf2 ->
	go_frame sl1 uf2
    | _::_ -> assert false (* Should never happen *) in
  let rec go sl1 cp2 = match sl1 with 
    | [] -> cp2
    | Csp_node(nid,e,cel)::sl1 -> go sl1 (wande_ptsto_cp (nid,e,cel) cp2)
    | (Csp_listsegi(SINGLE (snode,sfun,fld),e,f,_,_,ds) as s)::sl1 ->
	wande_lseg_cp (snode,e,f,ds) cp2
    | _ -> raise (Undef_wand_elim "Can only put pointsto on left of -*E") in
  (* Pick out all |-> that are allocated. *)
  let (sl1a,sl1b) = List.partition (function 
    | Csp_node(n,e,_) -> 
	List.exists (function Csp_node(m,x,_) -> Pervasives.(=) n m && equal_exp e x | _ -> false) (snd uf2)
    | _ -> false) sl1 in
  match go_frame sl1a uf2 with
    | None -> {cfl=[]}
    | Some uf2 -> go sl1b {cfl = [(uf2,scpl2)]}

(* returns  sl -*E cp *)
let rec wande_base_cp sl cp = 
  let cpl = List.map (wande_base_cform sl) cp.cfl in
  let cfl = List.concat (List.map (function cp -> cp.cfl) cpl) in
  {cfl = remove_can_form_duplicates cfl}

(* returns  cp1 -*E cp2 *)
let rec can_prop_wande cp1 cp2 =
  let rec f = function 
    | [] -> []
    | ((pl,sl),scpl)::xs -> 
	let cp2' = can_prop_star {cfl = [((pl,[]),scpl)]} cp2 in
	(wande_base_cp sl cp2') :: (f xs)
  in
  {cfl = List.concat (List.map (function cp -> cp.cfl) (f cp1.cfl))}


(** {2 Existential variables } *)


let ex_fv_can_prop cp =
  let fv = fv_can_prop cp in
  IdSet.filter is_existential fv

(** return the free existential variables contained in pointstos that mention TID .*)
let ex_fv_can_prop_fromTID cp =
  let fv = fv_can_prop_fromTID cp in
  IdSet.filter is_existential fv

(** Return substitutions [(sub_unex,sub_toex)] to turn the existential
    variables into normal variables, and the converse. *)
let remove_reinstate_existentials_sub cp =
  let exfv = ex_fv_can_prop cp in
  let il = IdSet.elements exfv in
  let iel_unex = List.map (fun id -> (id, Ce_ident (un_existential id))) il in
  let iel_toex = List.map (fun id -> ((un_existential id), Ce_ident id)) il
  in (mk_subst iel_unex, mk_subst iel_toex)

let fv_norm_can_prop cp =
  let fv = fv_can_prop cp
  in IdSet.filter (fun x -> not (is_existential x || is_garbage x)) fv

let fv_keep_can_spred sp = 
  let rec go res = function 
    | [] -> res
    | (c,e)::cel -> 
	if Config.is_value_field c || Config.is_lock_field c then go (fv_can_exp e ++ res) cel
	else go res cel in
  match sp with
  | Csp_node (_, ce, ccel) -> go (fv_can_exp ce) (fld_to_list ccel)
  | Csp_listsegi (_,ce1,ce2,ce3,ce4,el) ->
      (fv_can_exp ce1) ++ (fv_can_exp ce2) ++ (fv_can_exp ce3) ++ (fv_can_exp ce4) 
      ++ (IdSet.filter (fun x -> not (is_existential x)) (fv_dset el))
  | Csp_tree _ | Csp_indpred _ -> fv_can_spred sp

(* important variables not to be treated as garbage *)
let fv_keep_can_uform (pl,sl) =
  pure_fold (fun ca x -> fv_keep_can_atom ca ++ x) pl 
    (List.fold_left (fun s sp -> fv_keep_can_spred sp ++ s) IdSet.empty sl)

let rec fv_keep_can_form (uf,scpl) = 
  List.fold_left (fun s (_,cp') -> fv_keep_can_prop cp' ++ s) (fv_keep_can_uform uf) scpl

and fv_keep_can_prop cp =
  List.fold_left (fun s cf -> fv_keep_can_form cf ++ s) IdSet.empty cp.cfl

let rec fv_spatial_can_form ((_,sl),scpl) = 
  List.fold_left (fun s (_,cp') -> fv_spatial_can_prop cp' ++ s) (fv_can_uform (pure_true,sl)) scpl

and fv_spatial_can_prop cp =
  List.fold_left (fun s cf -> fv_spatial_can_form cf ++ s) IdSet.empty cp.cfl

let rec fv_more_keep_can_form (uf,scpl) = 
  IdSet.inter (List.fold_left (fun s (_,cp') -> fv_can_prop cp' ++ s) IdSet.empty scpl) (fv_can_uform (pure_true, snd uf))

let fv_killable_vars_form cf =
  let s = fv_can_form cf in
  let s1 = IdSet.filter (fun x->is_garbage x || is_existential x) s in
  let keep = IdSet.inter (fv_keep_can_form cf) (fv_spatial_can_form cf) in
  let kill = IdSet.diff s1 keep in
  let keep2 = fv_more_keep_can_form cf in
  let kill = IdSet.diff kill keep2 in
  kill

let fv_killable_vars cp =
  let s = fv_can_prop cp in
  List.fold_left (fun s1 cf -> IdSet.inter s1 (fv_killable_vars_form cf)) s cp.cfl
(*let fv_killable_vars cp =
  let s = fv_can_prop cp in
  let s1 = IdSet.filter (fun x->is_garbage x || is_existential x) s in
  let keep = IdSet.inter (fv_keep_can_prop cp) (fv_spatial_can_prop cp) in
  let kill = IdSet.diff s1 keep
  in kill*)

let kill_vars_can_uform (pl,sl) kill =
  let pl1 = pure_filter
    (fun ca -> IdSet.is_empty (IdSet.inter (fv_can_atom ca) kill)) pl in
  let sl1 = List.map
    (function Csp_node (s,e,cel) ->
           Csp_node (s,e,fld_filter (fun (c,e1) -> IdSet.is_empty (IdSet.inter (fv_can_exp e1) kill)) cel)
       | Csp_listsegi (k,e1,f1,e2,f2,el) ->
           let el' = List.filter (fun e' -> IdSet.is_empty (IdSet.inter (fv_can_exp e') kill)) (dset_to_list el) in
           Csp_listsegi (k,e1,f1,e2,f2,dset_from_list el')
       | csp -> csp) sl
  in (pl1,sl1)

let rec kill_vars_can_form (uf,scpl) kill =
 (kill_vars_can_uform uf kill, List.map (fun (s,cp')->(s,kill_vars_can_prop cp' kill)) scpl)

and kill_vars_can_prop cp kill =
  let rec f cp = 
    {cfl = List.map (fun cf -> kill_vars_can_form cf kill) cp.cfl}
  in f cp

let kill_garbage_vars (eq,cp) =
  let kill = fv_killable_vars cp in
  let (eq1,sub) = kill_vars_eq eq kill in
  let cp1 = kill_vars_can_prop (map_prop sub cp) kill
  in
  (* for debugging 
    if not (IdSet.is_empty kill)
    then
      fprintf !Config.formatter "@.STATE %a@.%a@.KILLING VARS %a@.RETURNS %a@.%a@."
	pp_eq eq pp_can_prop cp
	(pp_listsep pp_ident " " "") (IdSet.elements kill)
	pp_eq eq1 pp_can_prop cp1;
  *)
  (eq1,cp1)

let aggr_kill_garbage_vars cp =
  {cfl=List.map (fun cf -> kill_vars_can_form cf (fv_killable_vars_form cf)) cp.cfl}

(* for debug
let eq_add eq (e1,e2) =
  let _ = fprintf !Config.formatter "Extend equalities %a with %a=%a@."
      pp_eq eq pp_can_exp e1 pp_can_exp e2 in
  let eq' = eq_add eq (e1,e2) in
  let _ = fprintf !Config.formatter "Results in %a@." pp_eq eq'
  in eq'
*)

(** Return all free variables of [cp] not directly reachable from the set [s] *)
let fv_killable_vars2 s cp =
  let sr = ref s in
  let go_spred = function
    | Csp_node (_,ce, fld) ->
      if IdSet.for_all (fun x -> IdSet.mem x !sr) (fv_can_exp ce)  then  
        sr := !sr ++ fv_fld fld
    | _ -> () in
  let go_cprop cp = List.iter (fun((_,sl),_) -> List.iter go_spred sl) cp.cfl in
  let rec fix () =
     let s = !sr in
     go_cprop cp; 
     if IdSet.equal !sr s then () else fix () in
  fix ();
  IdSet.diff (fv_can_prop cp) !sr

let kill_dead_vars lv cp =
  let killfv = IdSet.elements (fv_killable_vars2 lv cp) in
  let sub = mk_subst (List.map (fun v -> (v,Ce_ident (Misc.into_existential v))) killfv) in
  normalize_can_prop (map_prop sub cp)

let neg_pure_list pl =
  let pl = remove_atom_duplicates pl in
  {cfl = List.map (fun ca -> ((pure_one (neg_can_atom ca),[]),[])) (pure_to_list pl)}
