(** Abstract syntax tree and operations for symbolic heaps. *)

open Misc
open Parsetree
open Exp

(** 
  This module defines:
    - the internal representation of assertions
    - constructors for such assertions
    - common operations (free variables, substitution, basic simplification)
    - interfaces to be used by the theorem prover and abstraction
    - abstract syntax tree for commands
*)

(** {2 Assertions} *)

(** Kind of link for linked lists *)
type link_kind = 
  | SINGLE of string * string * fields (* Node name, monoid name, common fields *)
  | DOUBLE of component * component 
  | XOR of component

(** Canonical heap predicate *)
type can_spred =
  | Csp_listsegi of link_kind * can_exp * can_exp * can_exp * can_exp * disj_set
  | Csp_node of string * can_exp * fields
  | Csp_tree of component * component * can_exp
  | Csp_indpred of string * can_exp list

(** Star-conjunction of heap predicates *)
type can_spatial = can_spred list

(** Canonical formulas without boxes *)
type can_uform = can_pure * can_spatial

(** Canonical formulas with boxes *)
type can_form = can_uform * (string * can_prop) list

(** Canonical proposition: disjunction of canonical formulas *)
and can_prop = {cfl: can_form list}

exception Undef_wand_elim of string
exception No_frame_found
exception Junk

(** {3 Operations on spatial formulas} *)

val spat_empty : can_spatial
val spat_one : can_spred -> can_spatial
val spat_star_one : can_spred -> can_spatial -> can_spatial
val spat_star : can_spatial -> can_spatial -> can_spatial
val spat_remove : can_spatial -> can_spred -> can_spatial
val spat_fold : (can_spred -> 'a -> 'a) -> can_spatial -> 'a -> 'a

val compare_can_spred : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_spred -> can_spred -> int

(** {3 Constructors} *)

val can_uform_false : can_uform
val can_form_false : can_form
val can_prop_false : can_prop

val can_prop_empty : can_prop
val is_can_prop_empty : can_prop -> bool
val can_atom_false: can_atom
val can_prop_pure: can_pure -> can_prop
val can_prop_spred: can_spatial -> can_prop

(** [can_form_star cf1 cf2] may return [None] if the result is inconsistent *)
val can_form_star: can_form -> can_form -> can_form option

val can_prop_star: can_prop -> can_prop -> can_prop
val and_cprop_pure : can_prop -> can_pure -> can_prop
val can_prop_or: can_prop -> can_prop -> can_prop

(** [can_prop_wande p q] septracts [p] from [q] (evaluates [p -*E q]).
  return [None] when the septraction is undefined.
 *)
val can_prop_wande: can_prop -> can_prop -> can_prop

(** {3 Comparisons} *)

val compare_uform : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_uform -> can_uform -> int
val compare_can_form : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_form -> can_form -> int
val compare_prop : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> can_prop -> can_prop -> int
val compare_string_prop : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> string * can_prop -> string * can_prop -> int
val compare_component_exp : ?cmp_id:(Misc.ident -> Misc.ident -> int) -> component * can_exp -> component * can_exp -> int

val remove_uform_duplicates : can_uform list -> can_uform list
val remove_can_form_duplicates : can_form list -> can_form list

val neg_pure_list : can_pure -> can_prop
val can_spred_leq : (Misc.ident * can_exp) list -> can_spred -> can_spred -> ((Misc.ident * can_exp) list * can_pure) option

(** {3 Free variables} *)

(** Return the free variables of an assertion *)
val fv_can_prop : can_prop -> Misc.IdSet.t

(** Return the non-existential free variables of an assertion *)
val fv_norm_can_prop : can_prop -> Misc.IdSet.t

(** Return the existential free variables of an assertion *)
val ex_fv_can_prop : can_prop -> Misc.IdSet.t

(** Return the existential free variables contained in pointstos that mention TID .*)
val ex_fv_can_prop_fromTID : can_prop -> Misc.IdSet.t

(** Returns existential free variables in a canonical order *)
val abs_fv_can_uform : can_uform -> Misc.ident list

(** {3 Substitutions} *)

val map_uform : (can_exp -> can_exp) -> can_uform -> can_uform option
val map_form : (can_exp -> can_exp) -> can_form -> can_form option
val map_prop : (can_exp -> can_exp) -> can_prop -> can_prop
val map_ip_body : (can_exp -> can_exp) -> (can_pure * can_form) list -> (can_pure * can_form) list
val map_idset : (can_exp -> can_exp) -> Misc.IdSet.t -> Misc.IdSet.t

(** Return substitutions [(sub_unex,sub_toex)] to turn the existential
    variables into normal variables, and the converse. *)
val remove_reinstate_existentials_sub : can_prop -> subst * subst

(** {3 Normalization} *)

(** Returns [None] if an inconsistency was detected. *)
val normalize_can_uform : can_uform -> can_uform option

(** Returns [None] if an inconsistency was detected. *)
val normalize_can_form : can_form -> can_form option

val normalize_can_prop : can_prop -> can_prop

val kill_garbage_vars : equalities * can_prop -> equalities * can_prop
val aggr_kill_garbage_vars : can_prop -> can_prop

(** Kill the variables of [cp] that are not directly reachable from the set [s] *)
val kill_dead_vars : Misc.IdSet.t -> can_prop -> can_prop


(** {2 Interfaces} *)

(** Interface provided by the theorem provers:
     - [find_ptsto] and [expose_ptsto]: used by symbolic execution, but should be removed.
     - [normal_form_prop sl_context eq cp = (eq',cp')] where [(eq',cp')] is a normal form of [(eq,cp)].
     - [normal_can_form cf] returns [None] if [cf] is inconsistent, 
       or [Some cf'] such that [cf <==> cf'] and [normal_can_form cf' = Some cf']
     - [entails_form cf cf'] returns [true] if [cf |- cf']
     - [find_frame_prop fr cp1 cp2] returns [res] such that [cp1 |- cp2 * res] in context [fr]
       (i.e. for all [R], [(fr /\ R) * cp1 |- (fr /\ R) * cp2 * res])
       or raises [No_frame_found]
 *)
type prover =
   { find_ptsto : string -> can_exp -> can_spatial -> fields * can_spatial;
     expose_ptsto : can_prop -> can_exp -> can_prop option;
     normal_form_prop : can_spatial -> equalities -> can_prop -> equalities * can_prop;
     normal_can_form : can_form -> can_form option;
     entails_form : can_form -> can_form -> bool;
     find_frame_prop : can_spatial -> can_prop -> can_prop -> can_prop;
   }

(** Interface provided by the abstraction module:
 - [uform_abs] is the analogue of [prop_abs] for formulas with no boxes. See below.
 - [uform_join] is the analogue of [prop_join] for formulas with no boxes. See below.
 - [prop_abs cp] returns a [cp'] belonging to a smaller domain such that [cp |- cp']
 - [prop_join cp1a cp1b cp2] assumes that [cp1a <==> cp1b] and returns 
      [(cpa,cpb,cp2')] such that [cpa <==> cpb] 
      and [cp1a |- cpa] and [cp2' |- cpa]
      and [(cp1b \/ cp2) ==> cpb] 
      and [cpb ==> (cp1b \/ cp2')]
  *)
type abstraction =
  {
     uform_abs : can_uform list -> can_uform list;
     uform_join : can_uform list -> can_uform list -> can_uform list -> (can_uform list * can_uform list * can_uform list);
     prop_abs : can_prop -> can_prop;
     prop_join : can_prop -> can_prop -> can_prop -> (can_prop * can_prop * can_prop);
  }


(** {2 Statements and commands} *)

(** Kind of branches *) 
type can_goto = Cgo_break | Cgo_continue | Cgo_return

(** Canonical statement *)
type can_stmt =
  { can_st_desc: can_st_desc      (** statement descriptor *) ;
    mutable can_st_lv: IdSet.t    (** set of live variables *);
    can_st_loc: Location.t        (** location for error messages *);
  }

(** Canonical command: sequence of statements *)
and can_cmd = can_stmt list

(** Statement descriptor *)
and can_st_desc =
  | Cst_nondet of can_cmd * can_cmd
  | Cst_assign of Misc.ident * can_exp
  | Cst_fldlookup of Misc.ident * can_exp * string
  | Cst_fldassign of can_exp * string * can_exp
  | Cst_new of Misc.ident
  | Cst_dispose of can_exp
  | Cst_pfcall of (ident * string * ident list * can_exp list) list
  | Cst_fcall2 of IdSet.t * can_prop * can_prop * string
  | Cst_assume of can_prop
  | Cst_region of can_cmd * string * Misc.IdSet.t * (string * can_prop * can_prop * can_prop) list
  | Cst_as_action of can_cmd * string * can_prop * can_prop * can_prop * Misc.IdSet.t
  | Cst_loop of can_cmd * can_prop option
  | Cst_goto of can_goto

(** Canonical entailments *)
type can_entailment = can_prop * can_cmd * can_prop * Location.t

type fun_info =
    { fun_param: ident list * ident list (** formal parameters (by reference, by value) *);
      fun_modif: IdSet.t        (** modified variables *);
      fun_pre: can_prop         (** precondition *);
      fun_post: can_prop        (** postcondition *);
      fun_loc: Location.t       (** location for error messages *) }

type res_init = 
  | Cri_inv of can_prop * (string * can_prop * can_prop * can_prop) list * Location.t
  | Cri_code of IdSet.t * can_cmd * (string * can_prop * can_prop * can_prop) list * Location.t

(** {2 Pretty printing} *)

val pp_listsep : (Format.formatter -> 'a -> unit) ->
      string -> string -> Format.formatter -> 'a list -> unit
val pp_can_spred : Format.formatter -> can_spred -> unit
val pp_can_form : Format.formatter -> can_form -> unit
val pp_can_prop : Format.formatter -> can_prop -> unit
val pp_can_cmd : Format.formatter -> can_cmd -> unit
val pp_can_entailment : Format.formatter -> can_prop * can_prop -> unit
val pp_inst : Format.formatter -> (Misc.ident * can_exp) list -> unit
